import pytest
from ase import Atoms
from gpaw import GPAW
import numpy as np
from gpaw.grid_descriptor import GridDescriptor
from gpaw.pw.descriptor import PWDescriptor
from gpaw.pw.poisson import ChargedReciprocalSpacePoissonSolver as CRSPC


def test_charged_pw_poisson():
    """Calculate Coulomb energy of Gaussian charge and compare to analytic
    result.
    """
    n = 50
    L = n / 40 * 8.0
    charge = 1.5
    gd = GridDescriptor((n, n, n), (L, L, L))
    pd = PWDescriptor(None, gd)
    ps = CRSPC(pd, charge, alpha=1.1)
    a = 0.8
    C = gd.cell_cv.sum(0) / 2
    rho = -np.exp(-1 / (4 * a) * pd.G2_qG[0] +
                  1j * (pd.get_reciprocal_vectors() @ C)) / gd.dv
    v = np.empty_like(rho)
    e = ps._solve(v, rho * charge)
    w = gd.collect(pd.ifft(v), broadcast=True)
    for pot, d in [(w[n // 2, n // 2, 0], L / 2),
                   (w[n // 2, 0, 0], L / 2 * 2**0.5),
                   (w[0, 0, 0], L / 2 * 3**0.5)]:
        assert pot == pytest.approx(-charge / d, abs=0.002)
    assert e == pytest.approx((a / 2 / np.pi)**0.5 * charge**2)


def test_pw_proton():
    """Check that the energy of a proton is 0.0."""
    proton = Atoms('H')
    proton.center(vacuum=2.0)
    proton.calc = GPAW(mode='pw', charge=1)
    e = proton.get_potential_energy()
    e += proton.calc.get_reference_energy()
    assert e == pytest.approx(0.0, abs=0.004)



import numpy as np
from ase.units import Bohr

from gpaw.fd_operators import Laplace, Gradient
from gpaw.kpoint import KPoint
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.lfc import LocalizedFunctionsCollection as LFC
from gpaw.mpi import serial_comm
from gpaw.preconditioner import Preconditioner
from gpaw.projections import Projections
from gpaw.transformers import Transformer
from gpaw.utilities.blas import axpy
from gpaw.wavefunctions.arrays import UniformGridWaveFunctions
from gpaw.wavefunctions.fdpw import FDPWWaveFunctions
from gpaw.wavefunctions.mode import Mode
import _gpaw


class FD(Mode):
    name = 'fd'

    def __init__(self, nn=3, interpolation=3, force_complex_dtype=False):
        self.nn = nn
        self.interpolation = interpolation
        Mode.__init__(self, force_complex_dtype)

    def __call__(self, *args, **kwargs):
        return FDWaveFunctions(self.nn, *args, **kwargs)

    def todict(self):
        dct = Mode.todict(self)
        dct['nn'] = self.nn
        dct['interpolation'] = self.interpolation
        return dct


class FDWaveFunctions(FDPWWaveFunctions):
    mode = 'fd'

    def __init__(self, stencil, parallel, initksl,
                 gd, nvalence, setups, bd,
                 dtype, world, kd, kptband_comm, timer, reuse_wfs_method=None,
                 collinear=True):
        FDPWWaveFunctions.__init__(self, parallel, initksl,
                                   reuse_wfs_method=reuse_wfs_method,
                                   collinear=collinear,
                                   gd=gd, nvalence=nvalence, setups=setups,
                                   bd=bd, dtype=dtype, world=world, kd=kd,
                                   kptband_comm=kptband_comm, timer=timer)

        # Kinetic energy operator:
        self.kin = Laplace(self.gd, -0.5, stencil, self.dtype)

        self.taugrad_v = None  # initialized by MGGA functional

    def empty(self, n=(), global_array=False, realspace=False, q=-1):
        return self.gd.empty(n, self.dtype, global_array)

    def integrate(self, a_xg, b_yg=None, global_integral=True):
        return self.gd.integrate(a_xg, b_yg, global_integral)

    def bytes_per_wave_function(self):
        return self.gd.bytecount(self.dtype)

    def set_setups(self, setups):
        self.pt = LFC(self.gd, [setup.pt_j for setup in setups],
                      self.kd, dtype=self.dtype, forces=True)
        FDPWWaveFunctions.set_setups(self, setups)

    def set_positions(self, spos_ac, atom_partition=None):
        FDPWWaveFunctions.set_positions(self, spos_ac, atom_partition)

    def __str__(self):
        s = 'Wave functions: Uniform real-space grid\n'
        s += '  Kinetic energy operator: %s\n' % self.kin.description
        return s + FDPWWaveFunctions.__str__(self)

    def make_preconditioner(self, block=1):
        return Preconditioner(self.gd, self.kin, self.dtype, block)

    def apply_pseudo_hamiltonian(self, kpt, ham, psit_xG, Htpsit_xG):
        self.timer.start('Apply hamiltonian')
        self.kin.apply(psit_xG, Htpsit_xG, kpt.phase_cd)
        ham.apply_local_potential(psit_xG, Htpsit_xG, kpt.s)
        ham.xc.apply_orbital_dependent_hamiltonian(
            kpt, psit_xG, Htpsit_xG, ham.dH_asp)
        self.timer.stop('Apply hamiltonian')

    def get_pseudo_partial_waves(self):
        phit_aj = [setup.get_partial_waves_for_atomic_orbitals()
                   for setup in self.setups]
        return LFC(self.gd, phit_aj, kd=self.kd, cut=True, dtype=self.dtype)

    def add_to_density_from_k_point_with_occupation(self, nt_sG, kpt, f_n):
        # Used in calculation of response part of GLLB-potential
        nt_G = nt_sG[kpt.s]
        for f, psit_G in zip(f_n, kpt.psit_nG):
            # Same as nt_G += f * abs(psit_G)**2, but much faster:
            _gpaw.add_to_density(f, psit_G, nt_G)

        # Hack used in delta-scf calculations:
        if hasattr(kpt, 'c_on'):
            assert self.bd.comm.size == 1
            d_nn = np.zeros((self.bd.mynbands, self.bd.mynbands),
                            dtype=complex)
            for ne, c_n in zip(kpt.ne_o, kpt.c_on):
                d_nn += ne * np.outer(c_n.conj(), c_n)
            for d_n, psi0_G in zip(d_nn, kpt.psit_nG):
                for d, psi_G in zip(d_n, kpt.psit_nG):
                    if abs(d) > 1.e-12:
                        nt_G += (psi0_G.conj() * d * psi_G).real

    def calculate_kinetic_energy_density(self):
        if self.taugrad_v is None:
            self.taugrad_v = [
                Gradient(self.gd, v, n=3, dtype=self.dtype).apply
                for v in range(3)]

        assert not hasattr(self.kpt_u[0], 'c_on')
        if not isinstance(self.kpt_u[0].psit_nG, np.ndarray):
            return None

        taut_sG = self.gd.zeros(self.nspins)
        dpsit_G = self.gd.empty(dtype=self.dtype)
        for kpt in self.kpt_u:
            for f, psit_G in zip(kpt.f_n, kpt.psit_nG):
                for v in range(3):
                    self.taugrad_v[v](psit_G, dpsit_G, kpt.phase_cd)
                    axpy(0.5 * f, abs(dpsit_G)**2, taut_sG[kpt.s])

        self.kptband_comm.sum(taut_sG)
        for taut_G in taut_sG:
            self.kd.symmetry.symmetrize(taut_G, self.gd)
        return taut_sG

    def apply_mgga_orbital_dependent_hamiltonian(self, kpt, psit_xG,
                                                 Htpsit_xG, dH_asp,
                                                 dedtaut_G):
        a_G = self.gd.empty(dtype=psit_xG.dtype)
        for psit_G, Htpsit_G in zip(psit_xG, Htpsit_xG):
            for v in range(3):
                self.taugrad_v[v](psit_G, a_G, kpt.phase_cd)
                self.taugrad_v[v](dedtaut_G * a_G, a_G, kpt.phase_cd)
                axpy(-0.5, a_G, Htpsit_G)

    def ibz2bz(self, atoms):
        """Transform wave functions in IBZ to the full BZ."""

        assert self.kd.comm.size == 1

        # New k-point descriptor for full BZ:
        kd = KPointDescriptor(self.kd.bzk_kc, nspins=self.nspins)
        kd.set_communicator(serial_comm)

        self.pt = LFC(self.gd, [setup.pt_j for setup in self.setups],
                      kd, dtype=self.dtype)
        self.pt.set_positions(atoms.get_scaled_positions())

        self.initialize_wave_functions_from_restart_file()

        weight = 2.0 / kd.nspins / kd.nbzkpts

        # Build new list of k-points:
        kpt_qs = []
        kpt_u = []
        for k in range(kd.nbzkpts):
            kpt_s = []
            for s in range(self.nspins):
                # Index of symmetry related point in the IBZ
                ik = self.kd.bz2ibz_k[k]
                r, q = self.kd.get_rank_and_index(ik)
                assert r == 0
                kpt = self.kpt_qs[q][s]

                phase_cd = np.exp(2j * np.pi * self.gd.sdisp_cd *
                                  kd.bzk_kc[k, :, np.newaxis])

                # New k-point:
                kpt2 = KPoint(1.0 / kd.nbzkpts, weight, s, k, k, phase_cd)
                kpt2.f_n = kpt.f_n / kpt.weight / kd.nbzkpts * 2 / self.nspins
                kpt2.eps_n = kpt.eps_n.copy()

                # Transform wave functions using symmetry operation:
                Psit_nG = self.gd.collect(kpt.psit_nG)
                if Psit_nG is not None:
                    Psit_nG = Psit_nG.copy()
                    for Psit_G in Psit_nG:
                        Psit_G[:] = self.kd.transform_wave_function(Psit_G, k)
                kpt2.psit = UniformGridWaveFunctions(
                    self.bd.nbands, self.gd, self.dtype,
                    kpt=k, dist=(self.bd.comm, self.bd.comm.size),
                    spin=kpt.s, collinear=True)
                self.gd.distribute(Psit_nG, kpt2.psit_nG)
                # Calculate PAW projections:
                nproj_a = [setup.ni for setup in self.setups]
                kpt2.projections = Projections(
                    self.bd.nbands, nproj_a,
                    kpt.projections.atom_partition,
                    self.bd.comm,
                    collinear=True, spin=s, dtype=self.dtype)

                kpt2.psit.matrix_elements(self.pt, out=kpt2.projections)
                kpt_s.append(kpt2)
                kpt_u.append(kpt2)
            kpt_qs.append(kpt_s)

        self.kd = kd
        self.kpt_qs = kpt_qs
        self.kpt_u = kpt_u

    def _get_wave_function_array(self, u, n, realspace=True, periodic=False):
        assert realspace
        kpt = self.kpt_u[u]
        psit_G = kpt.psit_nG[n]
        if periodic and self.dtype == complex:
            k_c = self.kd.ibzk_kc[kpt.k]
            return self.gd.plane_wave(-k_c) * psit_G
        return psit_G

    def write(self, writer, write_wave_functions=False):
        FDPWWaveFunctions.write(self, writer)

        if not write_wave_functions:
            return

        writer.add_array(
            'values',
            (self.nspins, self.kd.nibzkpts, self.bd.nbands) +
            tuple(self.gd.get_size_of_global_array()),
            self.dtype)

        for s in range(self.nspins):
            for k in range(self.kd.nibzkpts):
                for n in range(self.bd.nbands):
                    psit_G = self.get_wave_function_array(n, k, s)
                    writer.fill(psit_G * Bohr**-1.5)

    def read(self, reader):
        FDPWWaveFunctions.read(self, reader)

        if 'values' not in reader.wave_functions:
            return

        c = reader.bohr**1.5
        if reader.version < 0:
            c = 1  # old gpw file

        for kpt in self.kpt_u:
            # We may not be able to keep all the wave
            # functions in memory - so psit_nG will be a special type of
            # array that is really just a reference to a file:
            psit_nG = reader.wave_functions.proxy('values', kpt.s, kpt.k)
            psit_nG.scale = c

            kpt.psit = UniformGridWaveFunctions(
                self.bd.nbands, self.gd, self.dtype, psit_nG,
                kpt=kpt.q, dist=(self.bd.comm, self.bd.comm.size),
                spin=kpt.s, collinear=True)

        if self.world.size > 1:
            # Read to memory:
            for kpt in self.kpt_u:
                kpt.psit.read_from_file()

    def initialize_from_lcao_coefficients(self, basis_functions):
        for kpt in self.kpt_u:
            kpt.psit = UniformGridWaveFunctions(
                self.bd.nbands, self.gd, self.dtype, kpt=kpt.q,
                dist=(self.bd.comm, self.bd.comm.size, 1),
                spin=kpt.s, collinear=True)
            kpt.psit_nG[:] = 0.0
            mynbands = len(kpt.C_nM)
            basis_functions.lcao_to_grid(kpt.C_nM,
                                         kpt.psit_nG[:mynbands], kpt.q)
            kpt.C_nM = None

    def random_wave_functions(self, nao):
        """Generate random wave functions."""

        gpts = self.gd.N_c[0] * self.gd.N_c[1] * self.gd.N_c[2]

        if self.bd.nbands < gpts / 64:
            gd1 = self.gd.coarsen()
            gd2 = gd1.coarsen()

            psit_G1 = gd1.empty(dtype=self.dtype)
            psit_G2 = gd2.empty(dtype=self.dtype)

            interpolate2 = Transformer(gd2, gd1, 1, self.dtype).apply
            interpolate1 = Transformer(gd1, self.gd, 1, self.dtype).apply

            shape = tuple(gd2.n_c)
            scale = np.sqrt(12 / gd2.volume)

            old_state = np.random.get_state()

            np.random.seed(4 + self.world.rank)

            for kpt in self.kpt_u:
                for psit_G in kpt.psit_nG[nao:]:
                    if self.dtype == float:
                        psit_G2[:] = (np.random.random(shape) - 0.5) * scale
                    else:
                        psit_G2.real = (np.random.random(shape) - 0.5) * scale
                        psit_G2.imag = (np.random.random(shape) - 0.5) * scale

                    interpolate2(psit_G2, psit_G1, kpt.phase_cd)
                    interpolate1(psit_G1, psit_G, kpt.phase_cd)
            np.random.set_state(old_state)

        elif gpts / 64 <= self.bd.nbands < gpts / 8:
            gd1 = self.gd.coarsen()

            psit_G1 = gd1.empty(dtype=self.dtype)

            interpolate1 = Transformer(gd1, self.gd, 1, self.dtype).apply

            shape = tuple(gd1.n_c)
            scale = np.sqrt(12 / gd1.volume)

            old_state = np.random.get_state()

            np.random.seed(4 + self.world.rank)

            for kpt in self.kpt_u:
                for psit_G in kpt.psit_nG[nao:]:
                    if self.dtype == float:
                        psit_G1[:] = (np.random.random(shape) - 0.5) * scale
                    else:
                        psit_G1.real = (np.random.random(shape) - 0.5) * scale
                        psit_G1.imag = (np.random.random(shape) - 0.5) * scale

                    interpolate1(psit_G1, psit_G, kpt.phase_cd)
            np.random.set_state(old_state)

        else:
            shape = tuple(self.gd.n_c)
            scale = np.sqrt(12 / self.gd.volume)

            old_state = np.random.get_state()

            np.random.seed(4 + self.world.rank)

            for kpt in self.kpt_u:
                for psit_G in kpt.psit_nG[nao:]:
                    if self.dtype == float:
                        psit_G[:] = (np.random.random(shape) - 0.5) * scale
                    else:
                        psit_G.real = (np.random.random(shape) - 0.5) * scale
                        psit_G.imag = (np.random.random(shape) - 0.5) * scale

            np.random.set_state(old_state)

    def estimate_memory(self, mem):
        FDPWWaveFunctions.estimate_memory(self, mem)



"""GPAW command-line tool."""
import os
import subprocess
import sys


from ase.cli.main import main as ase_main

from gpaw import __version__


commands = [
    ('run', 'gpaw.cli.run'),
    ('info', 'gpaw.cli.info'),
    ('test', 'gpaw.cli.test'),
    ('dos', 'gpaw.cli.dos'),
    ('gpw', 'gpaw.cli.gpw'),
    ('completion', 'gpaw.cli.completion'),
    ('atom', 'gpaw.atom.aeatom'),
    ('diag', 'gpaw.fulldiag'),
    # ('quick', 'gpaw.cli.quick'),
    ('python', 'gpaw.cli.python'),
    ('sbatch', 'gpaw.cli.sbatch'),
    ('dataset', 'gpaw.atom.generator2'),
    ('symmetry', 'gpaw.symmetry'),
    ('install-data', 'gpaw.cli.install_data')]


def hook(parser, args):
    parser.add_argument('-P', '--parallel', type=int, metavar='N',
                        help='Run on N CPUs.')
    args = parser.parse_args(args)

    if args.command == 'python':
        args.traceback = True

    if hasattr(args, 'dry_run'):
        N = int(args.dry_run)
        if N:
            import gpaw
            gpaw.dry_run = N
            import gpaw.mpi as mpi
            mpi.world = mpi.SerialCommunicator()
            mpi.world.size = N

    if args.parallel:
        from gpaw.mpi import have_mpi, world
        if have_mpi and world.size == 1 and args.parallel > 1:
            py = sys.executable
        elif not have_mpi:
            py = 'gpaw-python'
        else:
            py = ''

        if py:
            # Start again in parallel:
            arguments = ['mpiexec', '-np', str(args.parallel), py]
            if args.command == 'python' and args.debug:
                arguments.append('-d')
            arguments += ['-m', 'gpaw']
            arguments += sys.argv[1:]

            extra = os.environ.get('GPAW_MPI_OPTIONS')
            if extra:
                arguments[1:1] = extra.split()

            # Use a clean set of environment variables without any MPI stuff:
            p = subprocess.run(arguments, check=not True, env=os.environ)
            sys.exit(p.returncode)

    return args


def main(args=None):
    ase_main('gpaw', 'GPAW command-line tool', __version__,
             commands, hook, args)



import optparse
import traceback
from pathlib import Path

import numpy as np
import ase.db
from ase import Atoms
from ase.data import covalent_radii, atomic_numbers
from ase.optimize import BFGS

from gpaw import GPAW, PW, KohnShamConvergenceError


cutoffs = [200, 250, 300, 400, 500, 600, 700, 800, 1500]


def check(con, name: str, lcao=True):
    params = dict(xc='PBE',
                  symmetry='off')

    if '.' in name:
        symbol, _, setup = name.partition('.')
        params['setups'] = setup
    else:
        symbol = name

    for h in [0.16, 0.18, 0.2]:
        a = 16 * h
        atoms = Atoms(symbol, cell=(a, a, 2 * a), pbc=True)
        atoms.calc = GPAW(h=h,
                          txt='{}-eggbox-{:.2f}.txt'.format(name, h),
                          **params)
        energies = []
        try:
            for i in range(4):
                energies.append(atoms.get_potential_energy())
                atoms.positions += h / 6
        except KohnShamConvergenceError:
            continue
        atoms.positions -= h / 6
        eegg = np.ptp(energies)
        con.write(atoms, name=name, test='eggbox', eegg=eegg, h=h)

    a = 4.0
    atoms = Atoms(symbol, cell=(a, a, a), pbc=True)
    for ecut in cutoffs:
        atoms.calc = GPAW(mode=PW(ecut),
                          txt='{}-pw-{:04}.txt'.format(name, ecut),
                          **params)
        try:
            atoms.get_potential_energy()
        except KohnShamConvergenceError:
            continue
        con.write(atoms, name=name, test='pw1', ecut=ecut)

    for g in [20, 24, 28]:
        atoms.calc = GPAW(gpts=(g, g, g),
                          txt='{}-fd-{}.txt'.format(name, g),
                          **params)
        try:
            atoms.get_potential_energy()
        except KohnShamConvergenceError:
            continue
        con.write(atoms, name=name, test='fd1', gpts=g)

    if lcao:
        for g in [20, 24, 28]:
            id = con.reserve(name=name, test='lcao1', gpts=g)
            if id is None:
                continue
            atoms.calc = GPAW(gpts=(g, g, g),
                              mode='lcao', basis='dzp',
                              txt='{}-lcao-{}.txt'.format(name, g),
                              **params)
            atoms.get_potential_energy()
            con.write(atoms, name=name, test='lcao1', gpts=g)
            del con[id]

    Z = atomic_numbers[symbol]
    d = 2 * covalent_radii[Z]
    atoms = Atoms(symbol * 2, cell=(a, a, 2 * a), pbc=True,
                  positions=[(0, 0, 0), (0, 0, d)])
    for ecut in cutoffs:
        atoms.calc = GPAW(mode=PW(ecut),
                          txt='{}2-pw-{:04}.txt'.format(name, ecut),
                          **params)
        try:
            atoms.get_potential_energy()
        except KohnShamConvergenceError:
            continue
        con.write(atoms, name=name, test='pw2', ecut=ecut)

    if 0:
        id = con.reserve(name=name, test='relax')
        if id is not None:
            atoms.calc = GPAW(mode=PW(1500),
                              txt='{}2-relax.txt'.format(name),
                              **params)
            BFGS(atoms).run(fmax=0.02)
            con.write(atoms, name=name, test='relax')
            del con[id]

    for g in [20, 24, 28]:
        atoms.calc = GPAW(gpts=(g, g, 2 * g),
                          txt='{}2-fd-{}.txt'.format(name, g),
                          **params)
        try:
            atoms.get_potential_energy()
        except KohnShamConvergenceError:
            continue
        con.write(atoms, name=name, test='fd2', gpts=g)

    if lcao:
        for g in [20, 24, 28]:
            id = con.reserve(name=name, test='lcao2', gpts=g)
            if id is None:
                continue
            atoms.calc = GPAW(gpts=(g, g, 2 * g),
                              mode='lcao', basis='dzp',
                              txt='{}2-lcao-{}.txt'.format(name, g),
                              **params)
            atoms.get_potential_energy()
            con.write(atoms, name=name, test='lcao2', gpts=g)
            del con[id]


def solve(energies, de):
    for i1 in range(len(energies) - 3, -1, -1):
        if energies[i1] > de:
            break
    c1 = cutoffs[i1]
    c2 = cutoffs[i1 + 1]
    # a * exp(-b * i)
    e1 = energies[i1]
    e2 = energies[i1 + 1]
    b = np.log(e1 / e2) / (c2 - c1)
    a = e1 * np.exp(b * c1)
    return np.log(a / de) / b


def summary(con, name):
    eegg = [row.get('eegg', np.nan)
            for row in con.select(name=name, test='eggbox', sort='h')]
    ecut = np.array([row.energy for row in con.select(name=name,
                                                      test='pw1',
                                                      sort='ecut')])
    ecut2 = np.array([row.get('energy', np.nan)
                      for row in con.select(name=name, test='pw2',
                                            sort='ecut')])
    eg = np.array([row.energy for row in con.select(name=name,
                                                    test='fd1', sort='gpts')])
    eg2 = np.array([row.get('energy', np.nan)
                    for row in con.select(name=name, test='fd2', sort='gpts')])
    eL = np.array([row.energy for row in con.select(name=name,
                                                    test='lcao1',
                                                    sort='gpts')])
    eL2 = np.array([row.get('energy', np.nan)
                    for row in con.select(name=name,
                                          test='lcao2',
                                          sort='gpts')])

    decut = ecut - 0.5 * ecut2
    energies = abs(ecut - ecut[-1])
    denergies = abs(decut - decut[-1])
    assert len(energies) == len(cutoffs)

    eg -= ecut[-1]
    eg2 -= ecut2[-1]
    deg = eg - 0.5 * eg2

    eL -= ecut[-1]
    eL2 -= ecut2[-1]
    deL = eL - 0.5 * eL2

    return (energies, denergies,
            abs(eg), abs(deg), abs(eL), abs(deL),
            eegg)


all_names = [
    'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Na.1',
    'Mg', 'Mg.2', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti',
    'V', 'V.5', 'Cr', 'Mn', 'Mn.7', 'Fe', 'Co', 'Ni', 'Ni.10', 'Cu', 'Zn',
    'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Nb.5',
    'Mo', 'Mo.6', 'Ru', 'Ru.8', 'Rh', 'Rh.9', 'Pd', 'Pd.10', 'Ag', 'Ag.11',
    'Cd', 'In', 'Sn', 'Sb', 'Te', 'Te.16', 'I', 'Xe', 'Cs', 'Ba', 'Hf',
    'Ta', 'Ta.5', 'W', 'W.6', 'Re', 'Os', 'Os.8', 'Ir', 'Ir.9', 'Pt',
    'Pt.10', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Rn']


def main():
    parser = optparse.OptionParser(usage='python -m gpaw.atom.check '
                                   '[options] symbol ...',
                                   description='Check dataset.')
    parser.add_option('-s', '--summary', action='store_true')
    parser.add_option('-v', '--verbose', action='store_true')
    parser.add_option('-p', '--plot', action='store_true')
    parser.add_option('-l', '--lcao', action='store_true')
    parser.add_option('-d', '--database', default='check.db')
    parser.add_option('--datasets', default='.')
    parser.add_option('-e', '--energy-difference', type=float, default=0.01)
    opts, names = parser.parse_args()
    if not names:
        names = [Path.cwd().name]
    con = ase.db.connect(opts.database)
    if opts.datasets:
        from gpaw import setup_paths
        setup_paths[:0] = opts.datasets.split(',')
    if opts.summary:
        for name in names:
            try:
                E, dE, eegg, ecut, decut, eg, deg, eL, deL = summary(
                    con, name, opts.energy_difference)
            except Exception as ex:
                if opts.verbose:
                    print(name)
                    traceback.print_exc()
                else:
                    print('{} {}: {}'.format(name,
                                             ex.__class__.__name__, ex))
            else:
                print('{0:5} {1:6.1f} {2:6.1f} {2:6.1f}'
                      .format(name, ecut, decut),
                      ''.join('{:7.4f}'.format(e) for e in eegg),
                      ''.join('{:7.3f}'.format(e) for e in eg),
                      ''.join('{:7.3f}'.format(e) for e in deg),
                      ''.join('{:7.3f}'.format(e) for e in eL),
                      ''.join('{:7.3f}'.format(e) for e in deL))
    if opts.plot:
        for name in names:
            E, dE, eegg, ecut, decut, eg, deg, eL, deL = summary(
                con, name, opts.energy_difference)
            import matplotlib.pyplot as plt
            plt.semilogy(cutoffs[:-1], E[:-1])
            plt.semilogy(cutoffs[:-1], dE[:-1])
            plt.semilogy([solve(E, de) for de in eg], eg, 's')
            plt.semilogy([solve(dE, de) for de in deg], deg, 'o')
            plt.semilogy([solve(E, de) for de in eL], eL, '+')
            plt.semilogy([solve(dE, de) for de in deL], deL, 'x')
        plt.show()
    else:
        for name in names:
            check(con, name, opts.lcao)


if __name__ == '__main__':
    main()



import os
import time

from gpaw import __version__


class FMF:
    """Full-Metadata Format

    Full-Metadata Format after
    http://www.sir.uni-freiburg.de/repository/2009/SI20090302a/SI20090302a.pdf
    """
    def __init__(self,
                 title='-',
                 creator=None,
                 place=None,
                 escape='#'):
        self.escape = escape
        self.estimate_creator(creator)
        self.title = title
        self.place = place

    def header(self):
        ec = self.escape
        header = ec + '; -*- fmf version: 1.0 -*-\n'
        header += ec + '[*reference]\n'
        header += ec + 'creator: ' + self.creator + '\n'
        header += ec + 'created: ' + time.strftime('%Y-%m-%d %H:%M') + '\n'

        title = self.title
        if isinstance(title, str):
            title = title.split('\n')
        for i, line in enumerate(title):
            if i == 0:
                header += ec + 'title: '
            else:
                header += ec + '       '
            header += line + '\n'

        header += ec + 'gpaw-version: ' + __version__ + '\n'
        u = os.uname()
        header += ec + 'hostname: ' + u.nodename + '\n'
        header += ec + 'architecture: ' + u.machine + '\n'

        return header

    def data(self, definitions):
        ec = self.escape
        data = ec + '[* data definitions]\n'
        for definition in definitions:
            data += ec + definition + '\n'
        data += ec + '[* data]\n'
        return data

    def field(self, title, entries):
        ec = self.escape
        res = ec + '[' + title + ']\n'
        for entry in entries:
            res += ec + entry + '\n'
        return res

    def estimate_creator(self, creator=None):
        self.creator = creator or os.environ.get('USER', 'unknown')



import numpy as np

from gpaw.pw.lfc import ft
from gpaw.gaunt import gaunt
from gpaw.spherical_harmonics import Y
from types import SimpleNamespace


class Setuplet:
    def __init__(self, *, phit_jg, phi_jg, rgd, l_j, rcut_j):
        self.rgd = rgd
        self.data = SimpleNamespace(phit_jg=phit_jg, phi_jg=phi_jg)
        self.l_j = l_j
        self.ni = np.sum([2 * l + 1 for l in l_j])
        self.rcut_j = rcut_j


def two_phi_planewave_integrals(qG_Gv, *, pawdata):
    rgd = pawdata.rgd
    l_j = pawdata.l_j
    phi_jg = pawdata.data.phi_jg
    phit_jg = pawdata.data.phit_jg
    ni = pawdata.ni
    
    gcut2 = rgd.ceil(2 * max(pawdata.rcut_j))
    
    # Initialize
    npw = qG_Gv.shape[0]
    phi_Gii = np.zeros((npw, ni, ni), dtype=complex)

    G_LLL = gaunt(max(l_j))
    k_G = np.sum(qG_Gv**2, axis=1)**0.5

    i1_start = 0

    for j1, l1 in enumerate(l_j):
        i2_start = 0
        for j2, l2 in enumerate(l_j):
            # Calculate the radial part of the product density
            rhot_g = phi_jg[j1] * phi_jg[j2] - phit_jg[j1] * phit_jg[j2]
            for l in range((l1 + l2) % 2, l1 + l2 + 1, 2):
                spline = rgd.spline(rhot_g[:gcut2], l=l, points=2**10)
                splineG = ft(spline, N=2**12)
                f_G = splineG.map(k_G) * (-1j)**l

                for m1 in range(2 * l1 + 1):
                    i1 = i1_start + m1
                    for m2 in range(2 * l2 + 1):
                        i2 = i2_start + m2
                        G_m = G_LLL[l1**2 + m1, l2**2 + m2, l**2:(l + 1)**2]
                        for m, G in enumerate(G_m):
                            # If Gaunt coefficient is zero, no need to add
                            if G == 0:
                                continue
                            x_G = Y(l**2 + m, *qG_Gv.T) * f_G
                            phi_Gii[:, i1, i2] += G * x_G

            i2_start += 2 * l2 + 1
        i1_start += 2 * l1 + 1
    return phi_Gii.reshape(npw, ni * ni)


class PWPAWCorrectionData:
    def __init__(self, Q_aGii, qpd, pawdatasets, pos_av):
        # Sometimes we loop over these in ways that are very dangerous.
        # It must be list, not dictionary.
        assert isinstance(Q_aGii, list)
        assert len(Q_aGii) == len(pos_av) == len(pawdatasets)

        self.Q_aGii = Q_aGii

        self.qpd = qpd
        self.pawdatasets = pawdatasets
        self.pos_av = pos_av

    def _new(self, Q_aGii):
        return PWPAWCorrectionData(Q_aGii, qpd=self.qpd,
                                   pawdatasets=self.pawdatasets,
                                   pos_av=self.pos_av)

    def remap(self, M_vv, G_Gv, sym, sign):
        Q_aGii = []
        for a, Q_Gii in enumerate(self.Q_aGii):
            x_G = self._get_x_G(G_Gv, M_vv, self.pos_av[a])
            U_ii = self.pawdatasets[a].R_sii[sym]

            Q_Gii = np.einsum('ij,kjl,ml->kim',
                              U_ii,
                              Q_Gii * x_G[:, None, None],
                              U_ii,
                              optimize='optimal')
            if sign == -1:
                Q_Gii = Q_Gii.conj()
            Q_aGii.append(Q_Gii)

        return self._new(Q_aGii)

    def _get_x_G(self, G_Gv, M_vv, pos_v):
        # This doesn't really belong here.  Or does it?  Maybe this formula
        # is only used with PAW corrections.
        return np.exp(1j * (G_Gv @ (pos_v - M_vv @ pos_v)))

    def remap_by_symop(self, symop, G_Gv, M_vv):
        return self.remap(M_vv, G_Gv, symop.symno, symop.sign)

    def multiply(self, P_ani, band):
        assert isinstance(P_ani, list)
        assert len(P_ani) == len(self.Q_aGii)

        C1_aGi = [Qa_Gii @ P1_ni[band].conj()
                  for Qa_Gii, P1_ni in zip(self.Q_aGii, P_ani)]
        return C1_aGi

    def reduce_ecut(self, G2G):
        # XXX actually we should return this with another PW descriptor.
        return self._new([Q_Gii.take(G2G, axis=0) for Q_Gii in self.Q_aGii])

    def almost_equal(self, otherpawcorr, G_G):
        for a, Q_Gii in enumerate(otherpawcorr.Q_aGii):
            e = abs(self.Q_aGii[a] - Q_Gii[G_G]).max()
            if e > 1e-12:
                return False
        return True


def get_pair_density_paw_corrections(pawdatasets, qpd, spos_ac):
    """Calculate and bundle paw corrections to the pair densities as a
    PWPAWCorrectionData object."""
    qG_Gv = qpd.get_reciprocal_vectors(add_q=True)
    pos_av = spos_ac @ qpd.gd.cell_cv

    # Collect integrals for all species:
    Q_xGii = {}
    for atom_index, pawdata in enumerate(pawdatasets):
        Q_Gii = two_phi_planewave_integrals(qG_Gv, pawdata=pawdata)
        ni = pawdata.ni
        Q_xGii[atom_index] = Q_Gii.reshape(-1, ni, ni)

    Q_aGii = []
    for atom_index, pawdata in enumerate(pawdatasets):
        Q_Gii = Q_xGii[atom_index]
        x_G = np.exp(-1j * (qG_Gv @ pos_av[atom_index]))
        Q_aGii.append(x_G[:, np.newaxis, np.newaxis] * Q_Gii)

    return PWPAWCorrectionData(Q_aGii, qpd=qpd,
                               pawdatasets=pawdatasets,
                               pos_av=pos_av)



from math import pi, acos, sqrt

import numpy as np
from ase.atoms import string2vector
from ase.units import Bohr, Hartree
from ase.utils import IOContext

from gpaw.spherical_harmonics import Y
from gpaw.utilities.tools import coordinates


class AngularIntegral:

    """Perform an angular integral on the grid.

    center:
      the center (Ang)
    gd:
      grid_descriptor of the grids to expand
    Rmax:
      maximal radius of the expansion (Ang)
    dR:
      grid spacing in the radius (Ang)
    """

    def __init__(self, center, gd, Rmax=None, dR=None):
        assert gd.orthogonal
        center = Vector3d(center) / Bohr

        self.center = center
        self.gd = gd

        # set Rmax to the maximal radius possible
        # i.e. the corner distance
        if not Rmax:
            Rmax = 0
            extreme = gd.h_cv.diagonal() * gd.N_c
            for corner in ([0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0],
                           [0, 0, 1], [1, 0, 1], [0, 1, 1], [1, 1, 1]):
                Rmax = max(Rmax,
                           self.center.distance(np.array(corner) * extreme))
        else:
            Rmax /= Bohr
        self.Rmax = Rmax

        if not dR:
            dR = min(gd.h_cv.diagonal())
        else:
            dR /= Bohr
        self.dR = dR

        self.initialize()

    def initialize(self):
        """Initialize grids."""

        Rmax = self.Rmax
        dR = self.dR
        gd = self.gd

        # initialize the ylm and Radial grids

        # self.V_R will contain the volume of the R shell
        # self.R_g will contain the radial indicees corresponding to
        #     each grid point
        # self.ball_g will contain the mask of the ball of radius Rmax
        # self.y_Lg will contain the YL values corresponding to
        #     each grid point
        V_R = np.empty((int(Rmax / dR + 1),))
        R_R = np.empty((int(Rmax / dR + 1),))

        r_cg, r2_g = coordinates(gd, self.center, tiny=1.e-78)
        r_g = np.sqrt(r2_g)
        rhat_cg = r_cg / r_g

        ball_g = np.where(r_g < Rmax, 1, 0)
        self.R_g = np.where(r_g < Rmax, r_g / dR, -1).astype(int)

        if hasattr(self, 'L_l'):  # ExpandYl
            npY = np.vectorize(Y, (float,), 'spherical harmonic')
            nL = len(self.L_l)
            y_Lg = []
            for L in range(nL):
                y_Lg.append(npY(L, rhat_cg[0], rhat_cg[1], rhat_cg[2]))
            self.y_Lg = y_Lg

        for i, v in enumerate(V_R):
            R_g = np.where(self.R_g == i, 1., 0.)
            V_R[i] = gd.integrate(R_g)
            R_R[i] = gd.integrate(R_g * r_g)

        self.ball_g = ball_g
        self.V_R = V_R
        self.nominalR_R = self.dR * (np.arange(len(self.V_R)) + .5)
        V_R = np.where(V_R > 0, V_R, -1)
        self.R_R = np.where(V_R > 0, R_R / V_R, self.nominalR_R)

    def integrate(self, f_g):
        """Integrate a function on the grid over the angles.

        Contains the weight 4*pi*R^2 with R in Angstrom."""
        int_R = []
        for i, dV in enumerate(self.V_R):
            # get the R shell
            R_g = np.where(self.R_g == i, 1, 0)
            int_R.append(self.gd.integrate(f_g * R_g) / self.dR)
        return np.array(int_R) * Bohr ** 2

    def average(self, f_g):
        """Give the angular average of a function on the grid."""
        V_R = np.where(self.V_R > 0, self.V_R, 1.e32)
        return self.integrate(f_g) * self.dR / V_R / Bohr ** 2

    def radii(self, model='nominal'):
        """Return the radii of the radial shells in Angstrom"""
        if model == 'nominal':
            return self.nominalR_R * Bohr
        elif model == 'mean':
            return self.R_R * Bohr
        else:
            raise NotImplementedError


class ExpandYl(AngularIntegral):

    """Expand the smooth wave functions in spherical harmonics
    relative to a given center.

    center:
      the center for the expansion (Ang)
    gd:
      grid_descriptor of the grids to expand
    lmax:
      maximal angular momentum in the expansion (lmax<7)
    Rmax:
      maximal radius of the expansion (Ang)
    dR:
      grid spacing in the radius (Ang)
    """

    def __init__(self, center, gd, lmax=6, Rmax=None, dR=None):

        self.lmax = lmax
        self.L_l = []
        for l in range(lmax + 1):
            for m in range(2 * l + 1):
                self.L_l.append(l)

        AngularIntegral.__init__(self, center, gd, Rmax, dR)

    def expand(self, psit_g):
        """Expand a wave function"""

        gamma_l = np.zeros((self.lmax + 1))
        nL = len(self.L_l)
        L_l = self.L_l

        def abs2(z):
            return z.real**2 + z.imag**2

        for i, dV in enumerate(self.V_R):
            # get the R shell and it's Volume
            R_g = np.where(self.R_g == i, 1, 0)
            if dV > 0:
                for L in range(nL):
                    psit_LR = self.gd.integrate(psit_g * R_g * self.y_Lg[L])
                    gamma_l[L_l[L]] += 4 * pi / dV * abs2(psit_LR)
        # weight of the wave function inside the ball
        weight = self.gd.integrate((psit_g * psit_g.conj()).real * self.ball_g)

        return gamma_l, weight

    def to_file(self, calculator,
                filename='expandyl.dat',
                spins=None,
                kpoints=None,
                bands=None
                ):
        """Expand a range of wave functions and write the result
        to a file"""
        with IOContext() as io:
            fd = io.openfile(filename)

            if not spins:
                srange = range(calculator.wfs.nspins)
            else:
                srange = spins
            if not kpoints:
                krange = range(len(calculator.wfs.kd.ibzk_kc))
            else:
                krange = kpoints
            if not bands:
                nrange = range(calculator.wfs.bd.nbands)
            else:
                nrange = bands

            print('# Yl expansion', 'of smooth wave functions', file=fd)
            lu = 'Angstrom'
            print('# center =', self.center * Bohr, lu, file=fd)
            print('# Rmax =', self.Rmax * Bohr, lu, file=fd)
            print('# dR =', self.dR * Bohr, lu, file=fd)
            print('# lmax =', self.lmax, file=fd)
            print('# s    k     n', end=' ', file=fd)
            print('kpt-wght    e[eV]      occ', end=' ', file=fd)
            print('    norm      sum   weight', end=' ', file=fd)
            spdfghi = 's p d f g h i'.split()
            for l in range(self.lmax + 1):
                print('      %' + spdfghi[l], end=' ', file=fd)
            print(file=fd)

            for s in srange:
                for k in krange:
                    u = k * calculator.wfs.nspins + s
                    for n in nrange:
                        kpt = calculator.wfs.kpt_u[u]
                        psit_G = kpt.psit_nG[n]
                        norm = self.gd.integrate((psit_G * psit_G.conj()).real)

                        gl, weight = self.expand(psit_G)
                        gsum = np.sum(gl)
                        gl = 100 * gl / gsum

                        print('%2d %5d %5d' % (s, k, n), end=' ', file=fd)
                        print('%6.4f %10.4f %8.4f' % (kpt.weight,
                                                      kpt.eps_n[n] * Hartree,
                                                      kpt.f_n[n]),
                              end=' ', file=fd)
                        print('%8.4f %8.4f %8.4f' %
                              (norm, gsum, weight), end=' ', file=fd)

                        for g in gl:
                            print('%8.2f' % g, end=' ', file=fd)
                        print(file=fd)
                        fd.flush()


class Vector3d(list):
    def __init__(self, vector=None):
        if vector is None:
            vector = [0, 0, 0]
        vector = string2vector(vector)
        list.__init__(self)
        for c in range(3):
            self.append(float(vector[c]))
        self.l = False

    def __add__(self, other):
        result = self.copy()
        for c in range(3):
            result[c] += other[c]
        return result

    def __truediv__(self, other):
        return Vector3d(np.array(self) / other)

    __div__ = __truediv__

    def __mul__(self, x):
        if isinstance(x, type(self)):
            return np.dot(self, x)
        else:
            return Vector3d(x * np.array(self))

    def __rmul__(self, x):
        return self.__mul__(x)

    def __lmul__(self, x):
        return self.__mul__(x)

    def __neg__(self):
        return -1 * self

    def __str__(self):
        return "(%g,%g,%g)" % tuple(self)

    def __sub__(self, other):
        result = self.copy()
        for c in range(3):
            result[c] -= other[c]
        return result

    def angle(self, other):
        """Return the angle between the directions of yourself and the
        other vector in radians."""
        other = Vector3d(other)
        ll = self.length() * other.length()
        if not ll > 0:
            return None
        return acos((self * other) / ll)

    def copy(self):
        return Vector3d(self)

    def distance(self, vector):
        if not isinstance(vector, type(self)):
            vector = Vector3d(vector)
        return (self - vector).length()

    def length(self, value=None):
        if value:
            fac = value / self.length()
            for c in range(3):
                self[c] *= fac
            self.l = False
        if not self.l:
            self.l = sqrt(self.norm())
        return self.l

    def norm(self):
        # return np.sum( self*self )
        return self * self  # XXX drop this class and use numpy arrays ...

    def x(self):
        return self[0]

    def y(self):
        return self[1]

    def z(self):
        return self[2]



import pytest
import numpy as np
from ase import Atoms

from gpaw import GPAW, PW, FermiDirac, Mixer
from gpaw.utilities import compiled_with_sl
from gpaw.response.df import DielectricFunction
from gpaw.mpi import world

# This test assures that some things that
# should be equal, are.


@pytest.mark.response
@pytest.mark.slow
def test_response_graphene(in_tmp_dir):
    a = 2.5
    c = 3.22

    GR = Atoms(symbols='C2',
               positions=[(0.5 * a, 0.2 - np.sqrt(3) / 6 * a, 0.0),
                          (0.5 * a, 0.2 + np.sqrt(3) / 6 * a, 0.0)],
               cell=[(0.5 * a, -0.5 * 3**0.5 * a, 0),
                     (0.5 * a, 0.5 * 3**0.5 * a, 0),
                     (0.0, 0.0, c * 2.0)])
    GR.set_pbc((True, True, True))
    atoms = GR
    # These are not physically-motivated kpoint grid sizes.
    # They are the (nearly) the minimum number we need to test
    # the symmetry off-on and gamma off-on difference.
    GSsettings = [
        {'symmetry': 'off', 'kpts': {'size': [3, 2, 1], 'gamma': False}},
        {'symmetry': {}, 'kpts': {'size': [3, 2, 1], 'gamma': False}},
        {'symmetry': 'off', 'kpts': {'size': [3, 2, 1], 'gamma': True}},
        {'symmetry': {}, 'kpts': {'size': [3, 2, 1], 'gamma': True}}]

    DFsettings = [{'disable_point_group': True,
                   'disable_time_reversal': True},
                  {'disable_point_group': False,
                   'disable_time_reversal': True},
                  {'disable_point_group': True,
                   'disable_time_reversal': False},
                  {'disable_point_group': False,
                   'disable_time_reversal': False}]

    if world.size > 1 and compiled_with_sl():
        DFsettings.append({'disable_point_group': False,
                           'disable_time_reversal': False,
                           'nblocks': 2})

    for GSkwargs in GSsettings:
        calc = GPAW(mode=PW(200),
                    occupations=FermiDirac(0.2),
                    mixer=Mixer(0.4),
                    convergence={'eigenstates': 1e-4, 'density': 1e-3},
                    **GSkwargs)

        atoms.calc = calc
        atoms.get_potential_energy()
        calc.write('gr.gpw', 'all')

        dfs = []
        for kwargs in DFsettings:
            DF = DielectricFunction(calc='gr.gpw',
                                    frequencies={'type': 'nonlinear',
                                                 'domega0': 0.2},
                                    eta=0.2,
                                    ecut=15.0,
                                    rate=0.001,
                                    **kwargs)
            df1, df2 = DF.get_dielectric_function()
            if world.rank == 0:
                dfs.append(df1)

        # Check the calculated dielectric functions against
        # each other.
        while len(dfs):
            df = dfs.pop()
            for DFkwargs, df2 in zip(DFsettings[-len(dfs):], dfs):
                try:
                    assert np.allclose(df, df2)
                    print('Ground state settings:', GSkwargs)
                    print('DFkwargs1:', DFsettings[-len(dfs) - 1])
                    print('DFkwargs2:', DFkwargs)
                    print(np.max(np.abs((df - df2) / df)))
                except AssertionError:
                    print('Some symmetry or block-par. related problems')
                    print('for calculation with following ground state')
                    print('settings')
                    print('Ground state settings:', GSkwargs)
                    print('The following DF settings do not return the ' +
                          'same results')
                    print('DFkwargs1:', DFsettings[-len(dfs) - 1])
                    print('DFkwargs2:', DFkwargs)
                    print(np.max(np.abs((df - df2) / df)))
                    raise AssertionError



import numpy as np
from ase.build import bulk
from gpaw import GPAW
from ase.dft.kpoints import monkhorst_pack


def test_ibzqpt():
    for kpt in (3, 4):
        kpts = (kpt, kpt, kpt)
        bzk_kc = monkhorst_pack(kpts)
        shift_c = []
        for Nk in kpts:
            if Nk % 2 == 0:
                shift_c.append(0.5 / Nk)
            else:
                shift_c.append(0.)
        bzk_kc += shift_c

        atoms = bulk('Si', 'diamond', a=5.431)
        calc = GPAW(h=0.2,
                    kpts=bzk_kc)

        atoms.calc = calc
        atoms.get_potential_energy()

        kd = calc.wfs.kd
        bzq_qc = kd.get_bz_q_points()
        ibzq_qc = kd.get_ibz_q_points(bzq_qc, calc.wfs.kd.symmetry.op_scc)[0]

        assert np.abs(bzq_qc - kd.bzk_kc).sum() < 1e-8
        assert np.abs(ibzq_qc - kd.ibzk_kc).sum() < 1e-8



from math import sqrt, pi

import numpy as np
from scipy.special import eval_legendre

from gpaw.xc.gga import (add_gradient_correction, gga_vars,
                         GGARadialExpansion, GGARadialCalculator,
                         get_gradient_ops)
from gpaw.xc.lda import calculate_paw_correction
from gpaw.xc.functional import XCFunctional
from gpaw.sphere.lebedev import weight_n


class MGGA(XCFunctional):
    orbital_dependent = True

    def __init__(self, kernel, stencil=2):
        """Meta GGA functional."""
        XCFunctional.__init__(self, kernel.name, kernel.type)
        self.kernel = kernel
        self.stencil = stencil
        self.fixed_ke = False

    def set_grid_descriptor(self, gd):
        self.grad_v = get_gradient_ops(gd, self.stencil)
        XCFunctional.set_grid_descriptor(self, gd)

    def get_setup_name(self):
        return 'PBE'

    # This method exists on GGA class as well.  Try to solve this
    # kind of problem when refactoring MGGAs one day.
    def get_description(self):
        return ('{} with {} nearest neighbor stencil'
                .format(self.name, self.stencil))

    def initialize(self, density, hamiltonian, wfs):
        self.wfs = wfs
        self.tauct = density.get_pseudo_core_kinetic_energy_density_lfc()
        self.tauct_G = None
        self.dedtaut_sG = None
        if ((not hasattr(hamiltonian, 'xc_redistributor'))
                or (hamiltonian.xc_redistributor is None)):
            self.restrict_and_collect = hamiltonian.restrict_and_collect
            self.distribute_and_interpolate = \
                density.distribute_and_interpolate
        else:
            def _distribute_and_interpolate(in_xR, out_xR=None):
                tmp_xR = density.interpolate(in_xR)
                if hamiltonian.xc_redistributor.enabled:
                    out_xR = hamiltonian.xc_redistributor.distribute(tmp_xR,
                                                                     out_xR)
                elif out_xR is None:
                    out_xR = tmp_xR
                else:
                    out_xR[:] = tmp_xR
                return out_xR

            def _restrict_and_collect(in_xR, out_xR=None):
                if hamiltonian.xc_redistributor.enabled:
                    in_xR = hamiltonian.xc_redistributor.collect(in_xR)
                return hamiltonian.restrict(in_xR, out_xR)
            self.restrict_and_collect = _restrict_and_collect
            self.distribute_and_interpolate = _distribute_and_interpolate

    def set_positions(self, spos_ac):
        self.tauct.set_positions(spos_ac, self.wfs.atom_partition)
        if self.tauct_G is None:
            self.tauct_G = self.wfs.gd.empty()
        self.tauct_G[:] = 0.0
        self.tauct.add(self.tauct_G)

    def calculate_impl(self, gd, n_sg, v_sg, e_g):
        sigma_xg, dedsigma_xg, gradn_svg = gga_vars(gd, self.grad_v, n_sg)
        self.process_mgga(e_g, n_sg, v_sg, sigma_xg, dedsigma_xg)
        add_gradient_correction(self.grad_v, gradn_svg, sigma_xg,
                                dedsigma_xg, v_sg)

    def fix_kinetic_energy_density(self, taut_sG):
        self.fixed_ke = True
        self._taut_gradv_init = False
        self._fixed_taut_sG = taut_sG.copy()

    def process_mgga(self, e_g, nt_sg, v_sg, sigma_xg, dedsigma_xg):
        if self.fixed_ke:
            taut_sG = self._fixed_taut_sG
            if not self._taut_gradv_init:
                self._taut_gradv_init = True
                # ensure initialization for calculation potential
                self.wfs.calculate_kinetic_energy_density()
        else:
            taut_sG = self.wfs.calculate_kinetic_energy_density()

        if taut_sG is None:
            taut_sG = self.wfs.gd.zeros(len(nt_sg))

        if 0:  # taut_sG is None:
            # Below code disabled because it produces garbage in at least
            # some cases.
            #
            # See https://gitlab.com/gpaw/gpaw/issues/124
            #
            # Initialize with von Weizsaecker kinetic energy density:
            nt0_sg = nt_sg.copy()
            nt0_sg[nt0_sg < 1e-10] = np.inf
            taut_sg = sigma_xg[::2] / 8 / nt0_sg
            nspins = self.wfs.nspins
            taut_sG = self.wfs.gd.empty(nspins)
            for taut_G, taut_g in zip(taut_sG, taut_sg):
                self.restrict_and_collect(taut_g, taut_G)
        else:
            taut_sg = np.empty_like(nt_sg)

        for taut_G, taut_g in zip(taut_sG, taut_sg):
            taut_G += 1.0 / self.wfs.nspins * self.tauct_G
            self.distribute_and_interpolate(taut_G, taut_g)

        # bad = taut_sg < tautW_sg + 1e-11
        # taut_sg[bad] = tautW_sg[bad]

        # m = 12.0
        # taut_sg = (taut_sg**m + (tautW_sg / 2)**m)**(1 / m)

        dedtaut_sg = np.empty_like(nt_sg)
        self.kernel.calculate(e_g, nt_sg, v_sg, sigma_xg, dedsigma_xg,
                              taut_sg, dedtaut_sg)

        self.dedtaut_sG = self.wfs.gd.empty(self.wfs.nspins)
        self.ekin = 0.0
        for s in range(self.wfs.nspins):
            self.restrict_and_collect(dedtaut_sg[s], self.dedtaut_sG[s])
            self.ekin -= self.wfs.gd.integrate(
                self.dedtaut_sG[s] * (taut_sG[s] -
                                      self.tauct_G / self.wfs.nspins))

    def stress_tensor_contribution(self, n_sg):
        sigma_xg, dedsigma_xg, gradn_svg = gga_vars(self.gd, self.grad_v, n_sg)
        taut_sG = self.wfs.calculate_kinetic_energy_density()
        if taut_sG is None:
            taut_sG = self.wfs.gd.zeros(len(n_sg))
        taut_sg = np.empty_like(n_sg)
        for taut_G, taut_g in zip(taut_sG, taut_sg):
            taut_G += 1.0 / self.wfs.nspins * self.tauct_G
            self.distribute_and_interpolate(taut_G, taut_g)

        nspins = len(n_sg)
        dedtaut_sg = np.empty_like(n_sg)
        v_sg = self.gd.zeros(nspins)
        e_g = self.gd.empty()
        self.kernel.calculate(e_g, n_sg, v_sg, sigma_xg, dedsigma_xg,
                              taut_sg, dedtaut_sg)

        def integrate(a1_g, a2_g=None):
            return self.gd.integrate(a1_g, a2_g, global_integral=False)

        P = integrate(e_g)
        for v_g, n_g in zip(v_sg, n_sg):
            P -= integrate(v_g, n_g)
        for sigma_g, dedsigma_g in zip(sigma_xg, dedsigma_xg):
            P -= 2 * integrate(sigma_g, dedsigma_g)
        for taut_g, dedtaut_g in zip(taut_sg, dedtaut_sg):
            P -= integrate(taut_g, dedtaut_g)

        tau_svvG = self.wfs.calculate_kinetic_energy_density_crossterms()

        stress_vv = P * np.eye(3)
        for v1 in range(3):
            for v2 in range(3):
                stress_vv[v1, v2] -= integrate(gradn_svg[0, v1] *
                                               gradn_svg[0, v2],
                                               dedsigma_xg[0]) * 2
                if nspins == 2:
                    stress_vv[v1, v2] -= integrate(gradn_svg[0, v1] *
                                                   gradn_svg[1, v2],
                                                   dedsigma_xg[1]) * 2
                    stress_vv[v1, v2] -= integrate(gradn_svg[1, v1] *
                                                   gradn_svg[1, v2],
                                                   dedsigma_xg[2]) * 2
        tau_cross_g = self.gd.empty()
        for s in range(nspins):
            for v1 in range(3):
                for v2 in range(3):
                    self.distribute_and_interpolate(
                        tau_svvG[s, v1, v2], tau_cross_g)
                    stress_vv[v1, v2] -= integrate(tau_cross_g, dedtaut_sg[s])

        self.dedtaut_sG = self.wfs.gd.empty(self.wfs.nspins)
        for s in range(self.wfs.nspins):
            self.restrict_and_collect(dedtaut_sg[s], self.dedtaut_sG[s])

        self.gd.comm.sum(stress_vv)
        return stress_vv

    def apply_orbital_dependent_hamiltonian(self, kpt, psit_xG,
                                            Htpsit_xG, dH_asp=None):
        self.wfs.apply_mgga_orbital_dependent_hamiltonian(
            kpt, psit_xG,
            Htpsit_xG, dH_asp,
            self.dedtaut_sG[kpt.s])

    def calculate_paw_correction(self, setup, D_sp, dEdD_sp=None,
                                 addcoredensity=True, a=None):
        assert not hasattr(self, 'D_sp')
        self.D_sp = D_sp
        self.n = 0
        self.ae = True
        self.xcc = setup.xc_correction
        self.dEdD_sp = dEdD_sp

        if self.xcc is not None and self.xcc.tau_npg is None:
            self.xcc.tau_npg, self.xcc.taut_npg = self.initialize_kinetic(
                self.xcc)

        rcalc = self.create_mgga_radial_calculator()
        expansion = GGARadialExpansion(rcalc)
        # The damn thing uses too many 'self' variables to define a clean
        # integrator object.
        E = calculate_paw_correction(expansion,
                                     setup, D_sp, dEdD_sp,
                                     addcoredensity, a)
        del self.D_sp, self.n, self.ae, self.xcc, self.dEdD_sp
        return E

    def create_mgga_radial_calculator(self):
        class MockKernel:
            def __init__(self, mgga):
                self.mgga = mgga

            def calculate(self, e_g, n_sg, v_sg, sigma_xg, dedsigma_xg):
                self.mgga.mgga_radial(e_g, n_sg, v_sg, sigma_xg, dedsigma_xg)

        return GGARadialCalculator(MockKernel(self))

    def mgga_radial(self, e_g, n_sg, v_sg, sigma_xg, dedsigma_xg):
        n = self.n
        nspins = len(n_sg)
        if self.ae:
            tau_pg = self.xcc.tau_npg[self.n]
            tauc_g = self.xcc.tauc_g / (sqrt(4 * pi) * nspins)
            sign = 1.0
        else:
            tau_pg = self.xcc.taut_npg[self.n]
            tauc_g = self.xcc.tauct_g / (sqrt(4 * pi) * nspins)
            sign = -1.0
        tau_sg = np.dot(self.D_sp, tau_pg) + tauc_g

        if 0:  # not self.ae:
            m = 12
            for tau_g, n_g, sigma_g in zip(tau_sg, n_sg, sigma_xg[::2]):
                tauw_g = sigma_g / 8 / n_g
                tau_g[:] = (tau_g**m + (tauw_g / 2)**m)**(1.0 / m)
                break

        dedtau_sg = np.empty_like(tau_sg)
        self.kernel.calculate(e_g, n_sg, v_sg, sigma_xg, dedsigma_xg,
                              tau_sg, dedtau_sg)
        if self.dEdD_sp is not None:
            self.dEdD_sp += (sign * weight_n[self.n] *
                             np.inner(dedtau_sg * self.xcc.rgd.dv_g, tau_pg))
        assert n == self.n
        self.n += 1
        if self.n == len(weight_n):
            self.n = 0
            self.ae = False

    def add_forces(self, F_av):
        dF_av = self.tauct.dict(derivative=True)
        self.tauct.derivative(self.dedtaut_sG.sum(0), dF_av)
        for a, dF_v in dF_av.items():
            F_av[a] += dF_v[0] / self.wfs.nspins

    def estimate_memory(self, mem):
        bytecount = self.wfs.gd.bytecount()
        mem.subnode('MGGA arrays', (1 + self.wfs.nspins) * bytecount)

    def initialize_kinetic(self, xcc):
        nii = xcc.nii
        nn = len(xcc.rnablaY_nLv)
        ng = len(xcc.phi_jg[0])

        tau_npg = np.zeros((nn, nii, ng))
        taut_npg = np.zeros((nn, nii, ng))
        create_kinetic(xcc, nn, xcc.phi_jg, tau_npg)
        create_kinetic(xcc, nn, xcc.phit_jg, taut_npg)
        return tau_npg, taut_npg


def create_kinetic(xcc, ny, phi_jg, tau_ypg):
    r"""Short title here.

    kinetic expression is::

                                         __         __
      tau_s = 1/2 Sum_{i1,i2} D(s,i1,i2) \/phi_i1 . \/phi_i2 +tauc_s

    here the orbital dependent part is calculated::

      __         __
      \/phi_i1 . \/phi_i2 =
                  __    __
                  \/YL1.\/YL2 phi_j1 phi_j2 +YL1 YL2 dphi_j1 dphi_j2
                                                     ------  ------
                                                       dr     dr
      __    __
      \/YL1.\/YL2 [y] = Sum_c A[L1,c,y] A[L2,c,y] / r**2

    """
    nj = len(phi_jg)
    dphidr_jg = np.zeros(np.shape(phi_jg))
    for j in range(nj):
        phi_g = phi_jg[j]
        xcc.rgd.derivative(phi_g, dphidr_jg[j])

    # Second term:
    for y in range(ny):
        i1 = 0
        p = 0
        Y_L = xcc.Y_nL[y]
        for j1, l1, L1 in xcc.jlL:
            for j2, l2, L2 in xcc.jlL[i1:]:
                c = Y_L[L1] * Y_L[L2]
                temp = c * dphidr_jg[j1] * dphidr_jg[j2]
                tau_ypg[y, p, :] += temp
                p += 1
            i1 += 1
    # first term
    for y in range(ny):
        i1 = 0
        p = 0
        rnablaY_Lv = xcc.rnablaY_nLv[y, :xcc.Lmax]
        Ax_L = rnablaY_Lv[:, 0]
        Ay_L = rnablaY_Lv[:, 1]
        Az_L = rnablaY_Lv[:, 2]
        for j1, l1, L1 in xcc.jlL:
            for j2, l2, L2 in xcc.jlL[i1:]:
                temp = (Ax_L[L1] * Ax_L[L2] + Ay_L[L1] * Ay_L[L2] +
                        Az_L[L1] * Az_L[L2])
                temp *= phi_jg[j1] * phi_jg[j2]
                temp[1:] /= xcc.rgd.r_g[1:]**2
                temp[0] = temp[1]
                tau_ypg[y, p, :] += temp
                p += 1
            i1 += 1
    tau_ypg *= 0.5


class PurePython2DMGGAKernel:
    def __init__(self, name, pars=None):
        self.name = name
        self.pars = pars
        self.type = 'MGGA'
        assert self.pars is not None

    def calculate(self, e_g, n_sg, dedn_sg,
                  sigma_xg, dedsigma_xg,
                  tau_sg, dedtau_sg):

        e_g[:] = 0.
        dedsigma_xg[:] = 0.
        dedtau_sg[:] = 0.

        # spin-paired:
        if len(n_sg) == 1:
            n = n_sg[0]
            n[n < 1e-20] = 1e-40
            sigma = sigma_xg[0]
            sigma[sigma < 1e-20] = 1e-40
            tau = tau_sg[0]
            tau[tau < 1e-20] = 1e-40

            # exchange
            e_x = twodexchange(n, sigma, tau, self.pars)
            e_g[:] += e_x * n

        # spin-polarized:
        else:
            n = n_sg
            n[n < 1e-20] = 1e-40
            sigma = sigma_xg
            sigma[sigma < 1e-20] = 1e-40
            tau = tau_sg
            tau[tau < 1e-20] = 1e-40

            # The spin polarized version is handle using the exact spin scaling
            # Ex[n1, n2] = (Ex[2*n1] + Ex[2*n2])/2
            na = 2.0 * n[0]
            nb = 2.0 * n[1]

            e2na_x = twodexchange(na, 4. * sigma[0], 2. * tau[0], self.pars)
            e2nb_x = twodexchange(nb, 4. * sigma[2], 2. * tau[1], self.pars)
            ea_x = e2na_x * na
            eb_x = e2nb_x * nb

            e_g[:] += (ea_x + eb_x) / 2.0


def twodexchange(n, sigma, tau, pars):
    # parameters for 2 Legendre polynomials
    parlen_i = int(pars[0])
    parlen_j = pars[2 + 2 * parlen_i]
    assert parlen_i == parlen_j
    pars_i = pars[1:2 + 2 * parlen_i]
    pars_j = pars[3 + 2 * parlen_i:]
    trans_i = pars_i[0]
    trans_j = pars_j[0]
    orders_i, coefs_i = np.split(pars_i[1:], 2)
    orders_j, coefs_j = np.split(pars_j[1:], 2)
    assert len(coefs_i) == len(orders_i)
    assert len(coefs_j) == len(orders_j)
    assert len(orders_i) == len(orders_j)

    # product Legendre expansion of Fx(s, alpha)
    e_x_ueg, rs = ueg_x(n)
    Fx = LegendreFx2(n, rs, sigma, tau,
                     trans_i, orders_i, coefs_i, trans_j, orders_j, coefs_j)
    return e_x_ueg * Fx


def LegendreFx2(n, rs, sigma, tau,
                trans_i, orders_i, coefs_i, trans_j, orders_j, coefs_j):
    # Legendre polynomial basis expansion in 2D

    # reduced density gradient in transformation t1(s)
    C2 = 0.26053088059892404
    s2 = sigma * (C2 * np.divide(rs, n))**2.
    x_i = transformation(s2, trans_i)
    assert x_i.all() >= -1.0 and x_i.all() <= 1.0

    # kinetic energy density parameter alpha in transformation t2(s)
    alpha = get_alpha(n, sigma, tau)
    x_j = transformation(alpha, trans_j)
    assert x_j.all() >= -1.0 and x_j.all() <= 1.0

    # product exchange enhancement factor
    Fx_i = legendre_polynomial(x_i, orders_i, coefs_i)
    # print(Fx_i);asdf
    Fx_j = legendre_polynomial(x_j, orders_j, coefs_j)
    Fx = Fx_i * Fx_j
    return Fx


def transformation(x, t):
    if t > 0:
        tmp = t + x
        x = 2.0 * np.divide(x, tmp) - 1.0
    elif int(t) == -1:
        tmp1 = (1.0 - x**2.0)**3.0
        tmp2 = (1.0 + x**3.0 + x**6.0)
        x = -1.0 * np.divide(tmp1, tmp2)
    else:
        raise KeyError('transformation %i unknown!' % t)
    return x


def get_alpha(n, sigma, tau):
    # tau LSDA
    aux = (3. / 10.) * (3.0 * np.pi * np.pi)**(2. / 3.)
    tau_lsda = aux * n**(5. / 3.)

    # von Weisaecker
    ind = (n != 0.).nonzero()
    gdms = np.maximum(sigma, 1e-40)  # |nabla rho|^2
    tau_w = np.zeros((np.shape(n)))
    tau_w[ind] = np.maximum(np.divide(gdms[ind], 8.0 * n[ind]), 1e-40)

    # z and alpha
    tau_ = np.maximum(tau_w, tau)
    alpha = np.divide(tau_ - tau_w, tau_lsda)
    assert alpha.all() >= 0.0
    return alpha


def ueg_x(n):
    C0I = 0.238732414637843
    C1 = -0.45816529328314287
    rs = (C0I / n)**(1 / 3.)
    ex = C1 / rs
    return ex, rs


def legendre_polynomial(x, orders, coefs):
    assert len(orders) == len(coefs) == 1
    return eval_legendre(orders[0], x) * coefs[0]



# Run with even number of cores

import pytest
from gpaw import GPAW, Mixer, PoissonSolver
import ase.parallel as mpi
from ase.build import fcc211, add_adsorbate
from ase.constraints import FixAtoms
from ase.optimize import QuasiNewton
from ase.neb import NEBtools
from ase.autoneb import AutoNEB


@pytest.mark.skip(reason='TODO')
def test_autoneb():
    size = mpi.world.size
    rank = mpi.world.rank

    slab = fcc211('Ag', size=(3, 2, 1), vacuum=3.0)
    add_adsorbate(slab, 'Ag', 0.5, (-0.1, 2.7))

    slab.set_constraint(FixAtoms(range(6)))
    slab.pbc = 1

    def getcalc(**kwargs):
        kwargs1 = dict(xc='oldLDA',
                       gpts=(32, 24, 32),
                       setups={'Ag': '11'},
                       # h=0.24,
                       poissonsolver=PoissonSolver(relax='GS'),
                       mixer=Mixer(0.5, 5, 50.0),
                       mode='lcao',
                       basis='sz(dzp)')
        kwargs1.update(kwargs)
        return GPAW(**kwargs1)

    calc = getcalc()

    slab.calc = calc

    qn = QuasiNewton(slab, trajectory='neb000.traj')
    qn.run(fmax=0.05)

    slab[-1].x += slab.get_cell()[0, 0]
    slab[-1].y += 2.8

    qn = QuasiNewton(slab, trajectory='neb001.traj')
    qn.run(fmax=0.05)

    def attach_calculators(images):
        nim = len(images)
        n = size // nim  # number of cpu's per image
        j = rank // n  # image number
        assert nim * n == size

        for i in range(nim):
            ranks = range(i * n, (i + 1) * n)
            if rank in ranks:
                calc = getcalc(txt='neb%d.txt' % j,
                               communicator=ranks)
                images[i].calc = calc

    autoneb = AutoNEB(attach_calculators,
                      prefix='neb',
                      n_simul=2,
                      parallel=True,
                      climb=True,
                      n_max=5,
                      optimizer='FIRE',
                      fmax=0.05,
                      k=0.5,
                      maxsteps=[25, 1000])
    autoneb.run()

    nebtools = NEBtools(autoneb.all_images)
    barrier, delta_e = nebtools.get_barrier()
    print('barrier', barrier)
    ref = 0.74051020956857272  # 1.484 <- with better parameters
    err = abs(barrier - ref)
    assert err < 1e-3, 'barrier={}, expected={}, err={}'.format(barrier,
                                                                ref, err)



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""Finite difference operators.

This file defines a series of finite difference operators used in grid mode.
"""

from math import pi, factorial as fact

import numpy as np
from numpy.fft import fftn, ifftn

import _gpaw
from gpaw import debug


# Expansion coefficients for finite difference Laplacian.  The numbers are
# from J. R. Chelikowsky et al., Phys. Rev. B 50, 11355 (1994):
laplace = [[0],
           [-2, 1],
           [-5 / 2, 4 / 3, -1 / 12],
           [-49 / 18, 3 / 2, -3 / 20, 1 / 90],
           [-205 / 72, 8 / 5, -1 / 5, 8 / 315, -1 / 560],
           [-5269 / 1800, 5 / 3, -5 / 21, 5 / 126, -5 / 1008, 1 / 3150],
           [-5369 / 1800, 12 / 7, -15 / 56, 10 / 189, -1 / 112, 2 / 1925,
            -1 / 16632]]

derivatives = [[1 / 2],
               [2 / 3, -1 / 12],
               [3 / 4, -3 / 20, 1 / 60],
               [4 / 5, -1 / 5, 4 / 105, -1 / 280]]


class FDOperator:
    def __init__(self, coef_p, offset_pc, gd, dtype=float,
                 description=None):
        """FDOperator(coefs, offsets, gd, dtype) -> FDOperator object.
        """

        # Is this a central finite-difference type of stencil?
        cfd = True
        for offset_c in offset_pc:
            if sum([offset != 0 for offset in offset_c]) >= 2:
                cfd = False

        mp = np.abs(offset_pc).max()  # padding

        # If stencil is strictly larger than any domain, some CPU will
        # need data from the second-nearest-neighbour domain.
        # We don't support that.  (Typically happens with 1D distributions)
        assert (mp <= gd.n_c).all(), 'Stencil longer than domain'
        n_c = gd.n_c
        M_c = n_c + 2 * mp
        stride_c = np.array([M_c[1] * M_c[2], M_c[2], 1])
        offset_p = np.dot(offset_pc, stride_c)
        coef_p = np.ascontiguousarray(coef_p, float)
        neighbor_cd = gd.neighbor_cd

        assert coef_p.ndim == 1
        assert coef_p.shape == offset_p.shape
        assert dtype in [float, complex]

        self.dtype = dtype
        self.shape = tuple(n_c)

        if gd.comm.size > 1:
            comm = gd.comm.get_c_object()
        else:
            comm = None

        assert neighbor_cd.flags.c_contiguous and offset_p.flags.c_contiguous
        self.mp = mp
        self.gd = gd
        self.npoints = len(coef_p)
        self.coef_p = coef_p
        self.offset_p = offset_p
        self.offset_pc = offset_pc
        self.comm = comm
        self.cfd = cfd

        self.operator = _gpaw.Operator(coef_p, offset_p, n_c, mp,
                                       neighbor_cd, dtype == float,
                                       comm, cfd)

        if description is None:
            description = '%d point finite-difference stencil' % self.npoints
        self.description = description

    def __str__(self):
        return '<' + self.description + '>'

    def apply(self, in_xg, out_xg, phase_cd=None):
        self.operator.apply(in_xg, out_xg, phase_cd)

    def relax(self, relax_method, f_g, s_g, n, w=None):
        self.operator.relax(relax_method, f_g, s_g, n, w)

    def get_diagonal_element(self):
        return self.operator.get_diagonal_element()

    def get_async_sizes(self):
        return self.operator.get_async_sizes()


if debug:
    _FDOperator = FDOperator

    class FDOperator(_FDOperator):  # type: ignore
        def apply(self, in_xg, out_xg, phase_cd=None):
            assert in_xg.shape == out_xg.shape
            assert in_xg.shape[-3:] == self.shape
            assert in_xg.flags.contiguous
            assert in_xg.dtype == self.dtype
            assert out_xg.flags.contiguous
            assert out_xg.dtype == self.dtype
            assert (self.dtype == float or
                    (phase_cd.dtype == complex and
                     phase_cd.shape == (3, 2)))
            _FDOperator.apply(self, in_xg, out_xg, phase_cd)

        def relax(self, relax_method, f_g, s_g, n, w=None):
            assert f_g.shape == self.shape
            assert s_g.shape == self.shape
            assert f_g.flags.contiguous
            assert f_g.dtype == float
            assert s_g.flags.contiguous
            assert s_g.dtype == float
            assert self.dtype == float
            _FDOperator.relax(self, relax_method, f_g, s_g, n, w)


def Laplace(gd, scale=1.0, n=1, dtype=float):
    if n == 9:
        return FTLaplace(gd, scale, dtype)
    else:
        return GUCLaplace(gd, scale, n, dtype)


class GUCLaplace(FDOperator):
    def __init__(self, gd, scale=1.0, n=1, dtype=float):
        """Laplacian for general non orthorhombic grid.

        gd: GridDescriptor
            Descriptor for grid.
        scale: float
            Scaling factor.  Use scale=-0.5 for a kinetic energy operator.
        n: int
            Range of stencil.  Stencil has O(h^(2n)) error.
        dtype: float or complex
            Data-type to work on.
        """

        # Order the 26 neighbor grid points after length
        # (reduced to 13 inequivalent):
        M_ic = np.indices((3, 3, 3)).reshape((3, -3)).T[-13:] - 1
        u_cv = gd.h_cv / (gd.h_cv**2).sum(1)[:, np.newaxis]**0.5
        u2_i = (np.dot(M_ic, u_cv)**2).sum(1)
        i_d = u2_i.argsort()

        # x^2, y^2, z^2, yz, xz, xy:
        m_mv = np.array([(2, 0, 0), (0, 2, 0), (0, 0, 2),
                         (0, 1, 1), (1, 0, 1), (1, 1, 0)])
        # Try 3, 4, 5 and 6 of the shortest directions:
        for D in range(3, 7):
            h_dv = np.dot(M_ic[i_d[:D]], gd.h_cv)
            A_md = (h_dv**m_mv[:, np.newaxis, :]).prod(2)
            # Find best stencil coefficients:
            a_d, residual, rank, s = np.linalg.lstsq(A_md, [1, 1, 1, 0, 0, 0],
                                                     rcond=-1)
            if residual.sum() < 1e-14:
                if rank != D:
                    raise ValueError(
                        'You have a weird unit cell!  '
                        'Try to use the maximally reduced Niggli cell.  '
                        'See the ase.build.niggli_reduce() function.')
                # D directions was OK
                break

        a_d *= scale
        offsets = [(0, 0, 0)]
        coefs = [laplace[n][0] * a_d.sum()]
        for d in range(D):
            M_c = M_ic[i_d[d]]
            offsets.extend(np.arange(1, n + 1)[:, np.newaxis] * M_c)
            coefs.extend(a_d[d] * np.array(laplace[n][1:]))
            offsets.extend(np.arange(-1, -n - 1, -1)[:, np.newaxis] * M_c)
            coefs.extend(a_d[d] * np.array(laplace[n][1:]))

        FDOperator.__init__(self, coefs, offsets, gd, dtype)

        self.description = (
            '%d*%d+1=%d point O(h^%d) finite-difference Laplacian' %
            ((self.npoints - 1) // n, n, self.npoints, 2 * n))


class Gradient(FDOperator):
    def __init__(self, gd, v, scale=1.0, n=1, dtype=float):
        """Symmetric gradient for general non orthorhombic grid.

        gd: GridDescriptor
            Descriptor for grid.
        v: int
            Direction of gradient: 0, 1, or 2 for x, y and z.
        scale: float
            Scaling factor.
        n: int
            Range of stencil.  Stencil has O(h^(2n)) error.
        dtype: float or complex
            Data-type to work on.
        """

        from scipy.spatial import Voronoi

        # Find nearest neighbors.  If h is a vector pointing at a
        # neighbor grid-points then we don't also include -h in the list:
        M_ic = np.indices((3, 3, 3)).reshape((3, -1)).T - 1
        h_iv = M_ic.dot(gd.h_cv)
        voro = Voronoi(h_iv)
        i_d = []  # List[int]
        for i1, i2 in voro.ridge_points:
            if i1 == 13 and i2 > 13:
                i_d.append(i2)
            elif i2 == 13 and i1 > 13:
                i_d.append(i1)

        D = len(i_d)  # number of neighbors (3, 4, 5, 6 or 7)

        h_dv = h_iv[i_d]  # vectors pointing at neighbor grid-points

        # Find gradient along 3 directions (n_cv):
        invh_vc = np.linalg.inv(gd.h_cv)
        n_cv = (invh_vc / (invh_vc**2).sum(axis=0)**0.5).T

        coef_cd = np.zeros((3, D))  # unknown weights
        for c, n_v in enumerate(n_cv):
            # Point on the plane perpendicular to the directions get a
            # weight of zero
            ok_d = abs(h_dv.dot(n_v)) > 1e-10
            h_mv = h_dv[ok_d]

            # The theee equations: A_jm.dot(coef_m) = [1, 0, 0]
            A_jm = np.array([h_mv.dot(n_v),
                             h_mv.dot(gd.h_cv[c - 1]),
                             h_mv.dot(gd.h_cv[c - 2])])

            U_jm, S_m, V_mm = np.linalg.svd(A_jm, full_matrices=False)
            coef_m = V_mm.T.dot(np.diag(S_m**-1).dot(U_jm.T.dot([1, 0, 0])))
            coef_cd[c, ok_d] = coef_m

        # Now get the v-gradient:
        coef_d = np.linalg.inv(n_cv)[v].dot(coef_cd) * scale

        # Create stencil:
        offsets = []
        coefs = []
        stencil = np.array(derivatives[n - 1])
        for d, c in enumerate(coef_d):
            if abs(c) < 1e-10:
                continue
            M_c = M_ic[i_d[d]]
            offsets.extend(np.arange(1, n + 1)[:, np.newaxis] * M_c)
            coefs.extend(c * stencil)
            offsets.extend(np.arange(-1, -n - 1, -1)[:, np.newaxis] * M_c)
            coefs.extend(-c * stencil)

        FDOperator.__init__(self, coefs, offsets, gd, dtype)

        self.description = (
            'Finite-difference {}-derivative with O(h^{}) error ({} points)'
            .format('xyz'[v], 2 * n, self.npoints))


class LaplaceA(FDOperator):
    def __init__(self, gd, scale, dtype=float):
        assert gd.orthogonal
        c = np.divide(-1 / 12, gd.h_cv.diagonal()**2) * scale  # Why divide?
        c0 = c[1] + c[2]
        c1 = c[0] + c[2]
        c2 = c[1] + c[0]
        a = -16.0 * np.sum(c)
        b = 10.0 * c + 0.125 * a
        FDOperator.__init__(self,
                            [a,
                             b[0], b[0],
                             b[1], b[1],
                             b[2], b[2],
                             c0, c0, c0, c0,
                             c1, c1, c1, c1,
                             c2, c2, c2, c2],
                            [(0, 0, 0),
                             (-1, 0, 0), (1, 0, 0),
                             (0, -1, 0), (0, 1, 0),
                             (0, 0, -1), (0, 0, 1),
                             (0, -1, -1), (0, -1, 1), (0, 1, -1), (0, 1, 1),
                             (-1, 0, -1), (-1, 0, 1), (1, 0, -1), (1, 0, 1),
                             (-1, -1, 0), (-1, 1, 0), (1, -1, 0), (1, 1, 0)],
                            gd, dtype,
                            'O(h^4) Mehrstellen Laplacian (A)')


class LaplaceB(FDOperator):
    def __init__(self, gd, dtype=float):
        a = 0.5
        b = 1.0 / 12.0
        FDOperator.__init__(self,
                            [a,
                             b, b, b, b, b, b],
                            [(0, 0, 0),
                             (-1, 0, 0), (1, 0, 0),
                             (0, -1, 0), (0, 1, 0),
                             (0, 0, -1), (0, 0, 1)],
                            gd, dtype,
                            'O(h^4) Mehrstellen Laplacian (B)')


class FTLaplace:
    def __init__(self, gd, scale, dtype):
        assert gd.comm.size == 1 and gd.pbc_c.all()

        N_c1 = gd.N_c[:, np.newaxis]
        i_cq = np.indices(gd.N_c).reshape((3, -1))
        i_cq += N_c1 // 2
        i_cq %= N_c1
        i_cq -= N_c1 // 2
        B_vc = 2.0 * pi * gd.icell_cv.T
        k_vq = np.dot(B_vc, i_cq)
        k_vq *= k_vq
        self.k2_Q = k_vq.sum(axis=0).reshape(gd.N_c)
        self.k2_Q *= -scale
        self.d = 6.0 / gd.h_cv[0, 0]**2
        self.npoints = 1000

    def apply(self, in_xg, out_xg, phase_cd=None):
        if in_xg.ndim > 3:
            for in_g, out_g in zip(in_xg, out_xg):
                out_g[:] = ifftn(fftn(in_g) * self.k2_Q).real
        else:
            out_xg[:] = ifftn(fftn(in_xg) * self.k2_Q).real

    def get_diagonal_element(self):
        return self.d


class OldGradient(FDOperator):
    def __init__(self, gd, v, scale=1.0, n=1, dtype=float):
        h = (gd.h_cv**2).sum(1)**0.5
        d = gd.xxxiucell_cv[:, v]
        A = np.zeros((2 * n + 1, 2 * n + 1))
        for i, io in enumerate(range(-n, n + 1)):
            for j in range(2 * n + 1):
                A[i, j] = io**j / float(fact(j))
        A[n, 0] = 1.0
        coefs = np.linalg.inv(A)[1]
        coefs = np.delete(coefs, len(coefs) // 2)
        offs = np.delete(np.arange(-n, n + 1), n)
        coef_p = []
        offset_pc = []
        for i in range(3):
            if abs(d[i]) > 1e-11:
                coef_p.extend(list(coefs * d[i] / h[i] * scale))
                offset = np.zeros((2 * n, 3), int)
                offset[:, i] = offs
                offset_pc.extend(offset)

        FDOperator.__init__(self, coef_p, offset_pc, gd, dtype,
                            'O(h^%d) %s-gradient stencil' % (2 * n, 'xyz'[v]))



"""
Calculate optical transition strengths.
"""

# General modules
import pytest
import numpy as np

# Script modules
from ase.build import bulk

from gpaw import GPAW, PW, FermiDirac
from gpaw.response.tool import (get_bz_transitions,
                                get_chi0_integrand,
                                get_degeneracy_matrix,
                                get_individual_transition_strengths)
from gpaw.test import equal


@pytest.mark.response
def test_response_pdens_tool(in_tmp_dir):
    # ------------------- Inputs ------------------- #

    # Part 1: ground state calculation
    pw = 200
    kpts = 3
    nbands = 8

    # Part 2: optical transitions calculation
    spins = 'all'
    q_c = [0., 0., 0.]
    bzk_kc = np.array([[0., 0., 0.]])

    # ------------------- Script ------------------- #

    # Part 1: ground state calculation
    a = 5.431
    atoms = bulk('Si', 'diamond', a=a)

    calc = GPAW(mode=PW(pw),
                kpts=(kpts, kpts, kpts),
                nbands=nbands,
                convergence={'bands': -1},
                xc='LDA',
                occupations=FermiDirac(0.001))  # use small FD smearing

    atoms.calc = calc
    atoms.get_potential_energy()  # get ground state density

    calc.write('si.gpw', 'all')  # write wavefunctions

    # Part 2: optical transition calculation

    pair, qpd, domainarg_td = get_bz_transitions('si.gpw', q_c, bzk_kc,
                                                 spins=spins,
                                                 ecut=10)

    nocc1, nocc2 = pair.gs.count_occupied_bands(1e-6)
    # XXX should we know 1e-6?

    # non-empty bands
    n_n = np.arange(0, nocc2)
    # not completely filled bands
    m_m = np.arange(nocc1, pair.gs.bd.nbands)

    nt = len(domainarg_td)
    nn = len(n_n)
    nm = len(m_m)
    nG = qpd.ngmax
    optical_limit = np.allclose(q_c, 0.)

    n_tnmG = np.zeros((nt, nn, nm, nG + 2 * optical_limit), dtype=complex)
    df_tnm = np.zeros((nt, nn, nm), dtype=float)
    eps_tn = np.zeros((nt, nn), dtype=float)
    eps_tm = np.zeros((nt, nm), dtype=float)
    for t, domainarg_d in enumerate(domainarg_td):
        (n_tnmG[t], df_tnm[t],
         eps_tn[t], eps_tm[t]) = get_chi0_integrand(pair, qpd,
                                                    n_n, m_m,
                                                    *domainarg_d)

    testNM_ibN = [[[0], [4, 5, 6]], [[0], [7]],
                  [[1, 2, 3], [4, 5, 6]], [[1, 2, 3], [7]]]
    testts_iG = np.array([[0.07, 0.07, 0.07], [0.00, 0.00, 0.00],
                          [51.34, 51.34, 51.34], [22.69, 22.69, 22.69]])

    for t, (domainarg_d, n_nmG,
            df_nm, eps_n, eps_m) in enumerate(zip(domainarg_td, n_tnmG, df_tnm,
                                                  eps_tn, eps_tm)):

        # Find degeneracies
        degmat_Nn, eps_N = get_degeneracy_matrix(eps_n, tol=1.e-2)
        degmat_Mm, eps_M = get_degeneracy_matrix(eps_m, tol=1.e-2)

        # Find diagonal transition strengths
        its_nmG = np.zeros((nn, nm, 1 + 2 * optical_limit))
        for G in range(1 + 2 * optical_limit):
            its_nmG[:, :, G] = get_individual_transition_strengths(
                n_nmG, df_nm,
                G, G)

        # Find unique transition strengths
        its_NmG = np.tensordot(degmat_Nn, its_nmG, axes=(1, 0))
        ts_MNG = np.tensordot(degmat_Mm, its_NmG, axes=(1, 1))
        ts_NMG = np.transpose(ts_MNG, (1, 0, 2))

        i = 0
        for N, ts_MG in enumerate(ts_NMG):
            for M, ts_G in enumerate(ts_MG):
                degN_n = n_n[np.where(degmat_Nn[N])]
                degM_m = m_m[np.where(degmat_Mm[M])]

                for testn, n in zip(testNM_ibN[i][0], degN_n):
                    equal(testn, n, 0.5)
                for testm, m in zip(testNM_ibN[i][1], degM_m):
                    equal(testm, m, 0.5)

                for testts, ts in zip(testts_iG[i], ts_G):
                    print(ts)
                    equal(testts, ts, 0.01)

                i += 1



from math import pi
from gpaw.core.plane_waves import PlaneWaves
from gpaw.core.pwacf import PlaneWaveAtomCenteredFunctions


class SpiralPW:
    def __init__(self,
                 pw: PlaneWaves,
                 qspiral_v):
        self.pw = pw
        self.qspiral_v = qspiral_v
        self.dtype = complex
        self.maxmysize = pw.maxmysize
        self.comm = pw.comm
        self.myshape = pw.myshape
        self.G_plus_k_Gv = pw.G_plus_k_Gv + qspiral_v
        self.ekin_G = 0.5 * (self.G_plus_k_Gv**2).sum(1)
        self.kpt = pw.kpt_c + pw.cell_cv @ qspiral_v / (2 * pi)
        self.kpt_c = self.kpt
        self.cell = pw.cell
        self.dv = pw.dv


class SpiralPWACF:
    def __init__(self, functions, positions, pw,
                 atomdist,
                 qspiral_v):
        self.pt_saiG = [
            PlaneWaveAtomCenteredFunctions(
                functions,
                positions,
                SpiralPW(pw, 0.5 * sign * qspiral_v),
                atomdist=atomdist)
            for sign in [1, -1]]

    def empty(self, dims, comm):
        return self.pt_saiG[0].empty(dims, comm)

    def integrate(self, psit_nsG, out):
        P_ansi = out
        for s, pt_aiG in enumerate(self.pt_saiG):
            pt_aiG._lazy_init()
            pt_aiG._lfc.integrate(psit_nsG.data[:, s],
                                  {a: P_nsi[:, s]
                                   for a, P_nsi in P_ansi.items()})

    def add_to(self, r_nsG, P_ansi):
        for s, pt_aiG in enumerate(self.pt_saiG):
            # pt_aiG._lazy_init()
            pt_aiG._lfc.add(r_nsG.data[:, s],
                            {a: P_nsi[:, s]
                             for a, P_nsi in P_ansi.items()})



from ase import Atoms
from gpaw import GPAW
from gpaw.tddft import TDDFT
from gpaw.tddft.ehrenfest import EhrenfestVelocityVerlet
from gpaw.test import equal
import pytest


@pytest.mark.later
def test_tddft_ehrenfest_nacl(in_tmp_dir):
    d = 4.0
    atoms = Atoms('NaCl', [(0, 0, 0), (0, 0, d)])
    atoms.center(vacuum=4.5)

    gs_calc = GPAW(nbands=4, eigensolver='cg', gpts=(32, 32, 44), xc='LDA',
                   setups={'Na': '1'})
    atoms.calc = gs_calc
    atoms.get_potential_energy()

    gs_calc.write('nacl_gs.gpw', 'all')

    td_calc = TDDFT('nacl_gs.gpw', propagator='EFSICN')
    evv = EhrenfestVelocityVerlet(td_calc, 0.001)

    i = 0
    evv.get_energy()
    r = evv.x[1][2] - evv.x[0][2]
    # print 'E = ', [i, r, evv.Etot, evv.Ekin, evv.e_coulomb]

    for i in range(5):
        evv.propagate(1.0)
        evv.get_energy()
        r = evv.x[1][2] - evv.x[0][2]
        print('E = ', [i + 1, r, evv.Etot, evv.Ekin, evv.e_coulomb])

    equal(r, 7.558883144, 1e-6)
    equal(evv.Etot, -0.10359175317017391, 1e-4)



from ase.build import molecule
from ase.parallel import parprint
from gpaw import GPAW
from gpaw.cluster import Cluster
from gpaw.test import equal


def test_fixocc():
    h = 0.4
    box = 2
    nbands = 2
    txt = '-'
    txt = None

    H2 = Cluster(molecule('H2'))
    H2.minimal_box(box, h)
    convergence = {'energy': 0.01, 'eigenstates': 0.001, 'density': 0.01}

    if 1:
        # test ZeroKelvin vs FixedOccupations
        c = GPAW(h=h, nbands=nbands,
                 occupations={'width': 0.0},
                 convergence=convergence,
                 txt=txt)
        H2.calc = c
        E_zk = H2.get_potential_energy()

        c = GPAW(h=h, nbands=nbands,
                 occupations=dict(name='fixed', numbers=[[1, 0]]),
                 convergence=convergence,
                 txt=txt)
        H2.calc = c
        E_fo = H2.get_potential_energy()
        parprint(E_zk, E_fo)
        equal(E_zk, E_fo, 1.e-10)

    if 1:
        # test spin-paired vs spin-polarized
        c = GPAW(h=h, nbands=nbands,
                 occupations={'name': 'fixed', 'numbers': [[0.5, 0.5]]},
                 convergence=convergence,
                 txt=txt)
        H2.calc = c
        E_ns = H2.get_potential_energy()
    if 1:
        c = GPAW(h=h, nbands=nbands, spinpol=True,
                 occupations={'name': 'fixed', 'numbers': [[0.5, 0.5]] * 2},
                 convergence=convergence,
                 txt=txt)
        H2.calc = c
        E_sp = H2.get_potential_energy()
        parprint(E_ns, E_sp)
        equal(E_ns, E_sp, 1.e-6)



import pytest
from gpaw.utilities import compiled_with_libvdwxc
import itertools
import numpy as np
from gpaw.grid_descriptor import GridDescriptor
from gpaw.xc import XC
from gpaw.xc.libvdwxc import vdw_df_cx
from gpaw.mpi import world

pytestmark = pytest.mark.skipif(not compiled_with_libvdwxc(),
                                reason='not compiled_with_libvdwxc()')


def test_vdw_libvdwxc_spin():
    def test(xc, tol=5e-10):
        N_c = np.array([10, 6, 4])
        # This test is totally serial.
        gd = GridDescriptor(N_c, N_c * 0.2, pbc_c=(1, 0, 1),
                            comm=world.new_communicator([world.rank]))
        actual_n = N_c - (1 - gd.pbc_c)

        gen = np.random.RandomState(17)
        n_sg = gd.zeros(2)
        n_sg[:] = gen.rand(*n_sg.shape)
        # sigma_xg = gd.zeros(3)
        # sigma_xg[:] = gen.random.rand(*sigma_xg.shape)

        if hasattr(xc, 'libvdwxc'):
            xc._nspins = 2
            xc.initialize_backend(gd)

        v_sg = gd.zeros(2)
        E = xc.calculate(gd, n_sg, v_sg)
        print('E', E)

        dn = 1e-6

        all_indices = itertools.product(range(2),
                                        range(1, actual_n[0], 2),
                                        range(0, actual_n[1], 2),
                                        range(0, actual_n[2], 2))

        for testindex in all_indices:
            n1_sg = n_sg.copy()
            n2_sg = n_sg.copy()
            v = v_sg[testindex] * gd.dv
            n1_sg[testindex] -= dn
            n2_sg[testindex] += dn

            E1 = xc.calculate(gd, n1_sg, v_sg.copy())
            E2 = xc.calculate(gd, n2_sg, v_sg.copy())

            dedn = 0.5 * (E2 - E1) / dn
            err = abs(dedn - v)
            print('{}{} v={} fd={} err={}'.format(xc.name, list(testindex),
                                                  v, dedn, err))
            assert err < tol, err

    test(XC('PBE'))
    test(vdw_df_cx())
    test(vdw_df_cx(vdwcoef=0.0))
    test(vdw_df_cx(vdwcoef=1e5), tol=2e-6)



import numpy as np
import pytest
from ase.build import molecule
from ase.parallel import parprint

from gpaw import GPAW
from gpaw.analyse.multipole import Multipole
from gpaw.cluster import Cluster
from gpaw.test import equal


@pytest.mark.later
def test_multipoleH2O(in_tmp_dir):
    h = 0.3

    s = Cluster(molecule('H2O'))
    s.minimal_box(3., h)

    gpwname = 'H2O_h' + str(h) + '.gpw'
    s.calc = GPAW(h=h, charge=0, txt=None)
    s.get_potential_energy()
    s.calc.write(gpwname)

    dipole_c = s.get_dipole_moment()
    parprint('Dipole', dipole_c)

    center = np.array([1, 1, 1]) * 50.
    mp = Multipole(center, s.calc, lmax=2)
    q_L = mp.expand(-s.calc.density.rhot_g)
    parprint('Multipole', q_L)

    # The dipole moment is independent of the center
    equal(dipole_c[2], q_L[2], 1e-10)

    mp.to_file(s.calc, mode='w')



import numpy as np
import pytest

from ase.parallel import world, parprint
from ase.units import Bohr
from gpaw import GPAW
from gpaw.lcao.dipoletransition import get_dipole_transitions
from gpaw.lrtddft.kssingle import KSSingles


@pytest.mark.later
def test_dipole_transition(gpw_files, tmp_path_factory):
    """Check dipole matrix-elements for Li."""
    calc = GPAW(gpw_files['bcc_li_lcao_wfs'], parallel=dict(sl_auto=True))
    # Initialize calculator if necessary
    if not hasattr(calc.wfs, 'C_nM'):
        calc.wfs.set_positions
        calc.initialize_positions(calc.atoms)
    from gpaw.kohnsham_layouts import BlacsOrbitalLayouts
    isblacs = isinstance(calc.wfs.ksl, BlacsOrbitalLayouts)  # XXX
    print(calc.wfs.ksl.using_blacs, isblacs)
    dip_skvnm = get_dipole_transitions(calc.wfs)
    parprint("Dipole moments calculated")
    assert dip_skvnm.shape == (1, 4, 3, 4, 4)
    dip_kvnm = dip_skvnm[0] * Bohr

    print(world.rank, dip_kvnm[:, 0, 0, 3])

    # check symmetry: abs(d[i,j]) == abs(d[j,i])
    for k in range(4):
        for v in range(3):
            dip_kvnm[k, v].T == pytest.approx(dip_kvnm[k, v])

    # Check numerical value of a few elements - signs might change!
    assert 0.0824 == pytest.approx(abs(dip_kvnm[0, 0, 0, 1]), abs=1e-4)
    assert abs(-0.0781 + 0.0282j) == pytest.approx(abs(dip_kvnm[1, 2, 0, 3]),
                                                   abs=1e-4)

    calc = GPAW(gpw_files['bcc_li_fd_wfs'])
    # compare to lrtddft implementation
    kss = KSSingles()
    atoms = calc.atoms
    atoms.calc = calc
    kss.calculate(calc.atoms, 0)
    lrrefv = []
    for ex in kss:
        print(ex)
        lrrefv.append(-1. * ex.muv * Bohr)
    lrrefv = np.array(lrrefv)

    # some printout for manual inspection, if wanted
    parprint("               lrtddft(fd)(v)        raman(v)")
    f = "{} {:+.4f}    {:+.4f}"
    # At gamma at three excited states are degenerate, so order is a problem
    parprint(f.format('k=0, 0->1 (x)', lrrefv[0, 0], dip_kvnm[0, 0, 0, 1]))
    parprint(f.format('k=0, 0->1 (y)', lrrefv[0, 1], dip_kvnm[0, 1, 0, 1]))
    parprint(f.format('k=0, 0->1 (z)', lrrefv[0, 2], dip_kvnm[0, 2, 0, 1]))
    parprint(f.format('k=0, 0->2 (x)', lrrefv[1, 0], dip_kvnm[0, 0, 0, 2]))
    parprint(f.format('k=0, 0->2 (y)', lrrefv[1, 1], dip_kvnm[0, 1, 0, 2]))
    parprint(f.format('k=0, 0->2 (z)', lrrefv[1, 2], dip_kvnm[0, 2, 0, 2]))
    parprint(f.format('k=0, 0->3 (x)', lrrefv[2, 0], dip_kvnm[0, 0, 0, 3]))
    parprint(f.format('k=0, 0->3 (y)', lrrefv[2, 1], dip_kvnm[0, 1, 0, 3]))
    parprint(f.format('k=0, 0->3 (z)', lrrefv[2, 2], dip_kvnm[0, 2, 0, 3]))
    parprint("")
    # At 2nd kpoint 2nd and 3rd excited state almost degenerate in LCAO,
    # order might be different from FD mode
    parprint(f.format('k=1, 0->1 (x)', lrrefv[3, 0], dip_kvnm[1, 0, 0, 1]))
    parprint(f.format('k=1, 0->1 (y)', lrrefv[3, 1], dip_kvnm[1, 1, 0, 1]))
    parprint(f.format('k=1, 0->1 (z)', lrrefv[3, 2], dip_kvnm[1, 2, 0, 1]))

    parprint(f.format('k=1, 0->2 (x)', lrrefv[4, 0], dip_kvnm[1, 0, 0, 2]))
    parprint(f.format('k=1, 0->2 (y)', lrrefv[4, 1], dip_kvnm[1, 1, 0, 2]))
    parprint(f.format('k=1, 0->2 (z)', lrrefv[4, 2], dip_kvnm[1, 2, 0, 2]))
    parprint(f.format('k=1, 0->3 (x)', lrrefv[5, 0], dip_kvnm[1, 0, 0, 3]))
    parprint(f.format('k=1, 0->3 (y)', lrrefv[5, 1], dip_kvnm[1, 1, 0, 3]))
    parprint(f.format('k=1, 0->3 (z)', lrrefv[5, 2], dip_kvnm[1, 2, 0, 3]))
    parprint("")
    # At the third k-point first and second excited state are degenerate
    parprint(f.format('k=2, 0->3 (x)', lrrefv[8, 0], dip_kvnm[2, 0, 0, 3]))
    parprint(f.format('k=2, 0->3 (y)', lrrefv[8, 1], dip_kvnm[2, 1, 0, 3]))
    parprint(f.format('k=2, 0->3 (z)', lrrefv[8, 2], dip_kvnm[2, 2, 0, 3]))
    # 4th k-point not included KSSingles, probably all bands above Fermi level



"""This module defines different external potentials."""
import copy
import warnings
from typing import Callable, Dict, Optional

import numpy as np
from ase.units import Bohr, Ha

import _gpaw
from gpaw.typing import Array3D

__all__ = ['ConstantPotential', 'ConstantElectricField', 'CDFTPotential',
           'PointChargePotential', 'StepPotentialz',
           'PotentialCollection']

known_potentials: Dict[str, Callable] = {}


def _register_known_potentials():
    from gpaw.bfield import BField
    known_potentials['CDFTPotential'] = lambda: None  # ???
    known_potentials['BField'] = BField
    for name in __all__:
        known_potentials[name] = globals()[name]


def create_external_potential(name, **kwargs):
    """Construct potential from dict."""
    if not known_potentials:
        _register_known_potentials()
    return known_potentials[name](**kwargs)


class ExternalPotential:
    vext_g: Optional[Array3D] = None
    vext_q: Optional[Array3D] = None

    def get_potential(self, gd):
        """Get the potential on a regular 3-d grid.

        Will only call calculate_potential() the first time."""

        if self.vext_g is None:
            self.calculate_potential(gd)
            self.vext_g.flags.writeable = False
        return self.vext_g

    def get_potentialq(self, gd, pd3):
        """Get the potential on a regular 3-d grid in real space.

        Will only call calculate_potential() the first time."""

        if self.vext_q is None:
            vext_g = self.get_potential(gd)
            self.vext_q = pd3.fft(vext_g)
            self.vext_q.flags.writeable = False

        return self.vext_q

    def calculate_potential(self, gd) -> None:
        raise NotImplementedError

    def get_name(self) -> str:
        return self.__class__.__name__

    def update_potential_pw(self, ham, dens) -> float:
        v_q = self.get_potentialq(ham.finegd, ham.pd3).copy()
        eext = ham.pd3.integrate(v_q, dens.rhot_q, global_integral=False)
        dens.map23.add_to1(ham.vt_Q, v_q)
        ham.vt_sG[:] = ham.pd2.ifft(ham.vt_Q)
        if not ham.collinear:
            ham.vt_xG[1:] = 0.0
        return eext

    def update_atomic_hamiltonians_pw(self, ham, W_aL, dens) -> None:
        vext_q = self.get_potentialq(ham.finegd, ham.pd3)
        dens.ghat.integrate(ham.vHt_q + vext_q, W_aL)

    def paw_correction(self, Delta_p, dH_sp) -> None:
        pass

    def derivative_pw(self, ham, ghat_aLv, dens) -> None:
        vext_q = self.get_potentialq(ham.finegd, ham.pd3)
        dens.ghat.derivative(ham.vHt_q + vext_q, ghat_aLv)


class NoExternalPotential(ExternalPotential):
    vext_g = np.zeros((0, 0, 0))

    def __str__(self):
        return 'NoExternalPotential'

    def update_potential_pw(self, ham, dens) -> float:
        ham.vt_sG[:] = ham.pd2.ifft(ham.vt_Q)
        if not ham.collinear:
            ham.vt_xG[1:] = 0.0
        return 0.0

    def update_atomic_hamiltonians_pw(self, ham, W_aL, dens):
        dens.ghat.integrate(ham.vHt_q, W_aL)

    def derivative_pw(self, ham, ghat_aLv, dens):
        dens.ghat.derivative(ham.vHt_q, ghat_aLv)


class ConstantPotential(ExternalPotential):
    """Constant potential for tests."""
    def __init__(self, constant=1.0):
        self.constant = constant / Ha
        self.name = 'ConstantPotential'

    def __str__(self):
        return 'Constant potential: {:.3f} V'.format(self.constant * Ha)

    def calculate_potential(self, gd):
        self.vext_g = gd.zeros() + self.constant

    def todict(self):
        return {'name': self.name,
                'constant': self.constant * Ha}


class ConstantElectricField(ExternalPotential):
    def __init__(self, strength, direction=[0, 0, 1], tolerance=1e-7):
        """External constant electric field.

        strength: float
            Field strength in V/Ang.
        direction: vector
            Polarization direction.
        """
        self.strength = strength * Bohr / Ha
        self.direction_v = np.array(direction, dtype=float)
        self.direction_v /= np.linalg.norm(self.direction_v)
        self.field_v = self.strength * self.direction_v
        self.tolerance = tolerance
        self.name = 'ConstantElectricField'

    def __str__(self):
        return ('Constant electric field: '
                '({:.3f}, {:.3f}, {:.3f}) V/Ang'
                .format(*(self.field_v * Ha / Bohr)))

    def calculate_potential(self, gd):
        # Note that PW-mode is periodic in all directions!
        L_c = abs(gd.cell_cv @ self.direction_v)
        # eps = self.tolerance
        # assert (L_c < eps).sum() == 2

        center_v = 0.5 * gd.cell_cv.sum(0)
        r_gv = gd.get_grid_point_coordinates().transpose((1, 2, 3, 0))
        f_g = (r_gv - center_v) @ self.direction_v

        # Set potential to zero at boundary of box (important for PW-mode).
        # Say we have 8 grid points.  Instead of a potential like this:
        #
        # -4 -3 -2 -1  0  1  2  3
        #
        # we want:
        #
        #  0 -3 -2 -1  0  1  2  3

        L = L_c.sum()
        f_g[abs(abs(f_g) - L / 2) < 1e-5] = 0.0

        self.vext_g = f_g * self.strength

    def todict(self):
        return {'name': self.name,
                'strength': self.strength * Ha / Bohr,
                'direction': self.direction_v}


class ProductPotential(ExternalPotential):
    def __init__(self, ext_i):
        self.ext_i = ext_i

    def calculate_potential(self, gd):
        self.vext_g = self.ext_i[0].get_potential(gd).copy()
        for ext in self.ext_i[1:]:
            self.vext_g *= ext.get_potential(gd)

    def __str__(self):
        return '\n'.join(['Product of potentials:'] +
                         [ext.__str__() for ext in self.ext_i])

    def todict(self):
        return {'name': self.__class__.__name__,
                'ext_i': [ext.todict() for ext in self.ext_i]}


class PointChargePotential(ExternalPotential):
    def __init__(self, charges, positions=None,
                 rc=0.2, rc2=np.inf, width=1.0):
        """Point-charge potential.

        charges: list of float
            Charges in units of `|e|`.
        positions: (N, 3)-shaped array-like of float
            Positions of charges in Angstrom.  Can be set later.
        rc: float
            Inner cutoff for Coulomb potential in Angstrom.
        rc2: float
            Outer cutoff for Coulomb potential in Angstrom.
        width: float
            Width for cutoff function for Coulomb part.

        For r < rc, 1 / r is replaced by a third order polynomial in r^2 that
        has matching value, first derivative, second derivative and integral.

        For rc2 - width < r < rc2, 1 / r is multiplied by a smooth cutoff
        function (a third order polynomium in r).

        You can also give rc a negative value.  In that case, this formula
        is used::

            (r^4 - rc^4) / (r^5 - |rc|^5)

        for all values of r - no cutoff at rc2!
        """
        self._dict = dict(name=self.__class__.__name__,
                          charges=charges, positions=positions,
                          rc=rc, rc2=rc2, width=width)
        self.q_p = np.ascontiguousarray(charges, float)
        self.rc = rc / Bohr
        self.rc2 = rc2 / Bohr
        self.width = width / Bohr
        if positions is not None:
            self.set_positions(positions)
        else:
            self.R_pv = None

        if abs(self.q_p).max() < 1e-14:
            warnings.warn('No charges!')
        if self.rc < 0. and self.rc2 < np.inf:
            warnings.warn('Long range cutoff chosen but will not be applied\
                           for negative inner cutoff values!')

    def todict(self):
        return copy.deepcopy(self._dict)

    def __str__(self):
        return ('Point-charge potential '
                '(points: {}, cutoffs: {:.3f}, {:.3f}, {:.3f} Ang)'
                .format(len(self.q_p),
                        self.rc * Bohr,
                        (self.rc2 - self.width) * Bohr,
                        self.rc2 * Bohr))

    def set_positions(self, R_pv, com_pv=None):
        """Update positions."""
        if com_pv is not None:
            self.com_pv = np.asarray(com_pv) / Bohr
        else:
            self.com_pv = None

        self.R_pv = np.asarray(R_pv) / Bohr
        self.vext_g = None

    def _molecule_distances(self, gd):
        if self.com_pv is not None:
            return self.com_pv - gd.cell_cv.sum(0) / 2

    def calculate_potential(self, gd):
        assert gd.orthogonal
        self.vext_g = gd.zeros()

        dcom_pv = self._molecule_distances(gd)

        _gpaw.pc_potential(gd.beg_c, gd.h_cv.diagonal().copy(),
                           self.q_p, self.R_pv,
                           self.rc, self.rc2, self.width,
                           self.vext_g, dcom_pv)

    def get_forces(self, calc):
        """Calculate forces from QM charge density on point-charges."""
        dens = calc.density
        F_pv = np.zeros_like(self.R_pv)
        gd = dens.finegd
        dcom_pv = self._molecule_distances(gd)

        _gpaw.pc_potential(gd.beg_c, gd.h_cv.diagonal().copy(),
                           self.q_p, self.R_pv,
                           self.rc, self.rc2, self.width,
                           self.vext_g, dcom_pv, dens.rhot_g, F_pv)
        gd.comm.sum(F_pv)
        return F_pv * Ha / Bohr


class CDFTPotential(ExternalPotential):
    # Dummy class to make cDFT compatible with new external
    # potential class ClassName(object):
    def __init__(self, regions, constraints, n_charge_regions,
                 difference):

        self.name = 'CDFTPotential'
        self.regions = regions
        self.constraints = constraints
        self.difference = difference
        self.n_charge_regions = n_charge_regions

    def todict(self):
        return {'name': 'CDFTPotential',
                # 'regions': self.indices_i,
                'constraints': self.v_i * Ha,
                'n_charge_regions': self.n_charge_regions,
                'difference': self.difference,
                'regions': self.regions}


class StepPotentialz(ExternalPotential):
    def __init__(self, zstep, value_left=0, value_right=0):
        """Step potential in z-direction

        zstep: float
            z-value that splits space into left and right [Angstrom]
        value_left: float
            Left side (z < zstep) potentential value [eV]. Default: 0
        value_right: float
            Right side (z >= zstep) potentential value [eV]. Default: 0
       """
        self.value_left = value_left
        self.value_right = value_right
        self.name = 'StepPotentialz'
        self.zstep = zstep

    def __str__(self):
        return 'Step potentialz: {0:.3f} V to {1:.3f} V at z={2}'.format(
            self.value_left, self.value_right, self.zstep)

    def calculate_potential(self, gd):
        r_vg = gd.get_grid_point_coordinates()
        self.vext_g = np.where(r_vg[2] < self.zstep / Bohr,
                               gd.zeros() + self.value_left / Ha,
                               gd.zeros() + self.value_right / Ha)

    def todict(self):
        return {'name': self.name,
                'value_left': self.value_left,
                'value_right': self.value_right,
                'zstep': self.zstep}


class PotentialCollection(ExternalPotential):
    def __init__(self, potentials):
        """Collection of external potentials to be applied

        potentials: list
            List of potentials
        """
        self.potentials = []
        for potential in potentials:
            if isinstance(potential, dict):
                potential = create_external_potential(
                    potential.pop('name'), **potential)
            self.potentials.append(potential)

    def __str__(self):
        text = 'PotentialCollection:\n'
        for pot in self.potentials:
            text += '  ' + pot.__str__() + '\n'
        return text

    def calculate_potential(self, gd):
        self.potentials[0].calculate_potential(gd)
        self.vext_g = self.potentials[0].vext_g.copy()
        for pot in self.potentials[1:]:
            pot.calculate_potential(gd)
            self.vext_g += pot.vext_g

    def todict(self):
        return {'name': 'PotentialCollection',
                'potentials': [pot.todict() for pot in self.potentials]}


def static_polarizability(atoms, strength=0.01):
    """Calculate polarizability tensor

    atoms: Atoms object
    strength: field strength in V/Ang

    Returns
    -------
    polarizability tensor:
        Unit (e^2 Angstrom^2 / eV).
        Multiply with Bohr * Ha to get (Angstrom^3)
    """
    atoms.get_potential_energy()
    calc = atoms.calc
    assert calc.parameters.external is None
    dipole_gs = calc.get_dipole_moment()

    alpha = np.zeros((3, 3))
    for c in range(3):
        axes = np.zeros(3)
        axes[c] = 1
        calc.set(external=ConstantElectricField(strength, axes))
        calc.get_potential_energy()
        alpha[c] = (calc.get_dipole_moment() - dipole_gs) / strength
    calc.set(external=None)

    return alpha.T



# Test that LCAO wavefunctions are available and equal after restarts
# in normal as well as 'all' mode

import pytest
import numpy as np
from ase.build import molecule
from gpaw import GPAW


@pytest.mark.skip(reason='TODO')
def test_restart():
    system = molecule('H2')
    system.center(vacuum=2.5)

    calc = GPAW(mode='lcao', basis='sz(dzp)', h=0.3, nbands=1, txt=None)
    system.calc = calc
    system.get_potential_energy()
    wf = calc.get_pseudo_wave_function(0)

    for mode in ['all', 'normal']:
        fname = 'lcao-restart.%s.gpw' % mode
        calc.write(fname, mode=dict(normal='', all='all')[mode])

        calc2 = GPAW(fname, txt=None)
        if mode == 'normal':
            continue
        wf2 = calc2.get_pseudo_wave_function(0)
        err = np.abs(wf2 - wf).max()
        print('%s: err=%s' % (mode, repr(err)))
        assert abs(err) < 1e-14



import pytest
from gpaw.mpi import world
from ase import Atoms
from gpaw import GPAW, PW
from gpaw.response.df import DielectricFunction
from gpaw.test import equal, findpeak

# Comparing the EELS spectrum of sodium for different block
# parallelizations. Intended to be run with 8 cores.
# Reasons that this can fail:
# - Bug in block parallelization

pytestmark = pytest.mark.skipif(world.size < 4, reason='world.size < 4')


@pytest.mark.response
@pytest.mark.slow
def test_response_na_plasmons(in_tmp_dir, scalapack):
    a = 4.23 / 2.0
    a1 = Atoms('Na',
               scaled_positions=[[0, 0, 0]],
               cell=(a, a, a),
               pbc=True)

    a1.calc = GPAW(mode=PW(300),
                   kpts={'size': (10, 10, 10), 'gamma': True},
                   parallel={'band': 1},
                   txt='small.txt')

    a1.get_potential_energy()
    a1.calc.diagonalize_full_hamiltonian(nbands=20)
    a1.calc.write('gs_Na.gpw', 'all')

    # Calculate the dielectric functions
    df1 = DielectricFunction('gs_Na.gpw',
                             nblocks=1,
                             ecut=40,
                             rate=0.001,
                             txt='1block.txt')

    df1NLFCx, df1LFCx = df1.get_dielectric_function(direction='x')

    df2 = DielectricFunction('gs_Na.gpw',
                             nblocks=4,
                             ecut=40,
                             rate=0.001,
                             txt='4block.txt')

    df2NLFCx, df2LFCx = df2.get_dielectric_function(direction='x')

    # Compare plasmon frequencies and intensities
    w_w = df1.wd.omega_w
    w1, I1 = findpeak(w_w, -(1. / df1LFCx).imag)
    w2, I2 = findpeak(w_w, -(1. / df2LFCx).imag)
    equal(w1, w2, 1e-2)
    equal(I1, I2, 1e-3)
